# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'custom_card',
    'label_tooltip_button',
    'navset_card_tab_jwst',
]

from typing import Optional
from htmltools import Tag
from shiny.ui._navs import (
    NavSet,
    NavSetCard,
)
from shiny.ui._card import card_body
from shiny._namespaces import resolve_id_or_none
import shiny.ui as ui


def custom_card(*args, body_args={}, **kwargs):
    """
    A wrapper over a Shiny card component with an explicit card_body
    (so that I can apply class_ and other customizations).
    """
    header = None
    args = list(args)
    for arg in args:
        # Only headers and footers are CardItems (right?)
        if isinstance(arg, ui.CardItem):
            header = arg
            args.remove(arg)
            break

    return ui.card(
        header,
        card_body(*args, **body_args),
        **kwargs,
    )



def label_tooltip_button(
        label, icons, tooltips, button_ids, placement='top',
        class_=None,
    ):
    """
    A label text which has one or more clickable icons (with tooltips).

    Parameters
    ----------
    label: String
        The label before the icons.
    icons: favicon.icon_svg() instance(s)
        This could be an icon_svg instance of a list of them.
        If it is a list, assume that tooltips and button_ids also are.
    tooltips: String(s)
        The tooltips for each icon.
    button_ids: String(s)
        The id for each button assigned to the icons.
    """
    if not isinstance(icons, list):
        icons = [icons]
        tooltips = [tooltips]
        button_ids = [button_ids]

    icon_buttons = [
        ui.tooltip(
            ui.input_action_link(
                id=button_id,
                label='',
                icon=icon,
            ),
            text,
            placement=placement,
        )
        for icon, text, button_id in zip(icons, tooltips, button_ids)
    ]
    return ui.div(
        label,
        *icon_buttons,
        class_=class_,
    )


class NavSetCardJWST(NavSet):
    """
    A NavSet look alike, but it doesn't really has multiple tab contents.
    """
    def __init__(
        self,
        *args,
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
        )

    def layout(self, nav: Tag, content: Tag) -> Tag:
        #print(nav)
        nav_style = ui.tags.style("""
            .navset-container {
                display: flex;
                width: 100%;
                flex-wrap: nowrap;
            }
            .navset-container .nav-item {
                flex-grow: 1;
                text-align: center;
                border-radius: 5px;
                margin: 1px;
                box-sizing: border-box;
                transition: border 0.3s ease;
                border: 1px solid transparent;
            }
            .navset-container .nav-item:hover {
                border: 1px solid #006cd4;
            }
        """)
        # Wrap nav items in a div with a specific class for styling
        return ui.TagList(
            nav_style,
            ui.div(nav, class_="navset-container"),
        )


def navset_card_tab_jwst(
        nav_panel_labels,
        id: Optional[str] = None,
        selected: Optional[str] = None,
    ) -> NavSetCard:
    """
    Render nav items as a pillset (buttons) but without the tabs-content.

    Parameters
    ----------
    nav_labels: List of string
        Labels to display for the nav buttons
    id: String
        input value that holds the currently selected nav item.
    selected: String
        Choose a particular nav item to select by default value
    """
    nav_panels = [
        ui.nav_panel(label, '')
        for label in nav_panel_labels
    ]
    return NavSetCardJWST(
        *nav_panels,
        ul_class="nav nav-pills",
        id=resolve_id_or_none(id),
        selected=selected,
    )



# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'ROOT',
    'KNOWN_PROGRAMS',
    'check_latest_version',
    'get_latest_pandeia_release',
    'get_version_advice',
    'get_pandeia_advice',
    'read_spectrum_file',
    'collect_spectra',
    'format_text',
    'pretty_print_target',
]

import os
from packaging.version import parse

from bs4 import BeautifulSoup
import numpy as np
import pyratbay.constants as pc
import pyratbay.tools as pt
import requests
from shiny import ui

ROOT = os.path.realpath(os.path.dirname(__file__)) + '/'
from .catalogs.utils import as_str


# Manually kept:
KNOWN_PROGRAMS = [
    1033, 1118, 1177, 1185, 1201, 1224, 1274, 1279, 1280, 1281, 1312,
    1331, 1353, 1366, 1442, 1541, 1633, 1729, 1743, 1803, 1846, 1935,
    1952, 1981, 2001, 2008, 2021, 2055, 2062, 2084, 2113, 2149, 2158,
    2159, 2304, 2319, 2334, 2347, 2358, 2372, 2420, 2437, 2454, 2488,
    2498, 2507, 2508, 2512, 2571, 2589, 2594, 2667, 2708, 2722, 2734,
    2759, 2765, 2783, 2950, 2961, 3077, 3154, 3171, 3231, 3235, 3263,
    3279, 3315, 3385, 3557, 3615, 3712, 3730, 3731, 3784, 3818, 3838,
    3860, 3942, 3969, 4008, 4082, 4098, 4102, 4105, 4126, 4195, 4227,
    4536, 4711, 4818, 4931, 5022, 5177, 5191, 5268, 5311, 5531, 5634,
    5687, 5799, 5844, 5863, 5866, 5882, 5894, 5924, 5959, 5967, 6045,
    6193, 6284, 6456, 6457, 6491, 6543,
    6932, 6978, 7073, 7188, 7251, 7255, 7407, 7675, 7683, 7686, 7849,
    7875, 7953, 7982, 8004, 8017, 8233, 8309, 8597, 8696, 8739, 8864,
    8877, 9025, 9033, 9095, 9101
]


def check_latest_version(package):
    response = requests.get(f'https://pypi.org/pypi/{package}/json')
    latest_version = response.json()['info']['version']
    return latest_version


def get_latest_pandeia_release():
    """
    Fetch latest pandeia.engine version for JWST from their website
    """
    url = 'https://outerspace.stsci.edu/display/PEN/Pandeia+Engine+News'
    try:
        response = requests.get(url)
        status_code = response.status_code
    except:
        status_code = 0
    if status_code == 200:
        soup = BeautifulSoup(response.text, 'html.parser')
        for a in soup.find_all('a', href=True):
            link = 'pypi.org/project/pandeia.engine/'
            if link in a['href']:
                href = a['href']
                ini = href.index(link) + len(link)
                version = href[ini:]
                if '/' in version:
                    version = version[0:version.index('/')]
                return version.strip()
    # hard-coded default, need to be kept up to date manually:
    return '2025.3'


def get_version_advice(package, latest_version=None):
    name = package.__name__
    my_version = parse(package.__version__)
    if latest_version is None:
        latest_version = parse(check_latest_version(name))
    else:
        latest_version = parse(latest_version)
    my_major_minor = parse(f'{my_version.major}.{my_version.minor}')
    latest_major_minor = parse(f'{latest_version.major}.{latest_version.minor}')
    if my_version == latest_version:
        color = '#0B980D'
        advice = ''
    elif my_major_minor == latest_major_minor:
        color = '#ffa500'
        advice = (
            f'.<br>You may want to upgrade {name} with:<br>'
            f'<span style="font-weight:bold;">pip install --upgrade {name}</span>'
        )
    else:
        color = 'red'
        advice = (
            f'.<br>You should upgrade {name} with:<br>'
            f'<span style="font-weight:bold;">pip install --upgrade {name}</span>'
        )

    status_advice = ui.HTML(
        f'<br><p><span style="color:{color}">You have {name} '
        f'version {my_version}, the latest version is '
        f'{latest_version}</span>{advice}</p>'
    )
    return status_advice


def get_pandeia_advice(package, latest_version):
    name = package.__name__
    my_version = parse(package.__version__)
    latest_version = parse(latest_version)
    if my_version >= latest_version:
        color = '#0B980D'
        advice = ''
    else:
        color = 'red'
        advice = (
            f'.<br>You should upgrade {name} with:<br>'
            f'<span style="font-weight:bold;">pip install --upgrade {name}</span>'
        )

    status_advice = ui.HTML(
        f'<br><p><span style="color:{color}">You have {name} '
        f'version {my_version}, the latest JWST version is '
        f'{latest_version}</span>{advice}</p>'
    )
    return status_advice


def read_spectrum_file(file, units='none', on_fail=None):
    """
    Parameters
    ----------
    file: String
        Spectrum file to read (transit depth, eclipse depth, or stellar SED)
        This is a plain-text file with two columns (white space separater)
        First column is the wavelength, second is the depth/flux.
        Should be readable by numpy.loadtxt().
    units: String
        Units of the input spectrum.
        For depths, use one of 'none', 'percent', 'ppm'.
        For SEDs, use one of
            'mJy', 
            'f_freq' (for erg s-1 cm-2 Hz-1),
            'f_nu' (for for erg s-1 cm-2 cm),
            'f_lambda' (for erg s-1 cm-2 cm-1)
    on_fail: String
        if 'warning' raise a warning.
        if 'error' raise an error.

    Examples
    --------
    >>> import gen_tso.utils as u

    >>> file = f'{u.ROOT}data/models/WASP80b_transit.dat'
    >>> label, wl, spectra = u.read_spectrum_file(file, on_fail='warning')
    """
    # Validate units
    depth_units = [
        "none",
        "percent",
        "ppm",
    ]
    sed_units = [
        "f_freq",
        "f_nu",
        "f_lambda",
        "mJy",
    ]
    if units not in depth_units and units not in sed_units:
        msg = (
            f"The input units ({repr(units)}) must be one of {depth_units} "
            f"for depths or one of {sed_units} for SEDs"
        )
        raise ValueError(msg)

    # (try to) load the file:
    try:
        wl, spectrum = np.loadtxt(file, unpack=True)
    except ValueError as error:
        error_msg = (
            f'Error, could not load spectrum file: {repr(file)}\n{error}'
        )
        if on_fail == 'warning':
            print(error_msg)
        if on_fail == 'error':
            raise ValueError(error_msg)
        return None, None, None

    # Set the units:
    if units in depth_units:
        u = pt.u(units)
    else:
        if units == 'f_freq':
            u = 10**26
        elif units == 'f_nu':
            u = 10**26 / pc.c
        elif units == 'f_lambda':
            u = 10**26 / pc.c * (wl*pc.um)**2.0
        elif 'mJy' in units:
            u = 1.0

    path, label = os.path.split(file)
    if label.endswith('.dat') or label.endswith('.txt'):
        label = label[0:-4]
    return label, wl, spectrum*u


def collect_spectra(folder, on_fail=None):
    """
    Collect transit, eclipse, and SED spectra files from folder.
    - Transit spectra are identified by having 'transmission' or 'transit'
    in their names.  Or all files contained in a 'transit/' subfolder.
    - Eclipse spectra are identified by having 'emission' or 'eclipse'
    in their names.  Or all files contained in a 'eclipse/' subfolder.
    - SED spectra are identified by having 'sed' or 'star' in their names.

    Parameters
    ----------
    folder: String
        The folder where to search.
    on_fail: String
        if 'warning' raise a warning.
        if 'error' raise an error.

    Returns
    -------
    transit_spectra: Dictionary
        Spectrum name: 1D spectrum pairs for each transit file.
            The 1D spectra is itself a dictionary with keys: wl and depth.
    eclipse_spectra: Dictionary
        Spectrum name: 1D spectrum pairs for each eclipse file.
            The 1D spectra is itself a dictionary with keys: wl and depth.
    sed_spectra: Dictionary
        Spectrum name: 1D spectrum pairs for each SED file.
            The 1D spectra is itself a dictionary with keys: wl and flux.

    Examples
    --------
    >>> import gen_tso.utils as u

    >>> folder = f'{u.ROOT}data/models/'
    >>> spectra = u.collect_spectra(folder, on_fail=None)
    """
    files = sorted(os.listdir(folder))
    transit_files = [
        file for file in files
        if 'transit' in file or 'transmission' in file
        if not os.path.isdir(f'{folder}/{file}')
    ]
    if 'transit' in files and os.path.isdir(f'{folder}/transit'):
        sub_folder = f'{folder}/transit'
        transit_files += [
            f'transit/{file}' for file in sorted(os.listdir(sub_folder))
            if not os.path.isdir(f'{folder}/transit/{file}')
        ]

    eclipse_files = [
        file for file in files
        if 'eclipse' in file or 'emission' in file
        if not os.path.isdir(f'{folder}/{file}')
    ]
    if 'eclipse' in files and os.path.isdir(f'{folder}/eclipse'):
        sub_folder = f'{folder}/eclipse'
        eclipse_files += [
            f'eclipse/{file}' for file in sorted(os.listdir(sub_folder))
            if not os.path.isdir(f'{folder}/eclipse/{file}')
        ]

    sed_files = [
        file for file in files
        if 'sed' in file or 'star' in file
        if not os.path.isdir(f'{folder}/{file}')
    ]
    if 'sed' in files and os.path.isdir(f'{folder}/sed'):
        sub_folder = f'{folder}/sed'
        sed_files += [
            f'sed/{file}' for file in sorted(os.listdir(sub_folder))
            if not os.path.isdir(f'{folder}/sed/{file}')
        ]

    transit_spectra = {}
    for file in transit_files:
        filename = f'{folder}/{file}'
        units = 'none'
        label, wl, depth = read_spectrum_file(filename, units, on_fail)
        if wl is not None:
            transit_spectra[label] = {
                'wl': wl,
                'depth': depth,
                'units': units,
                'filename': filename,
            }

    eclipse_spectra = {}
    for file in eclipse_files:
        filename = f'{folder}/{file}'
        units = 'none'
        label, wl, depth = read_spectrum_file(filename, units, on_fail)
        if wl is not None:
            eclipse_spectra[label] = {
                'wl': wl,
                'depth': depth,
                'units': units,
                'filename': filename,
            }

    sed_spectra = {}
    for file in sed_files:
        filename = f'{folder}/{file}'
        units = 'mJy'
        label, wl, model = read_spectrum_file(filename, units, on_fail)
        if wl is not None:
            sed_spectra[label] = {
                'wl': wl,
                'flux': model,
                'units': units,
                'filename': filename,
            }

    return transit_spectra, eclipse_spectra, sed_spectra


def format_text(text, warning=False, danger=False, format=None):
    """
    Return a colorful text depending on requested format and warning
    or danger flags.

    Parameters
    ----------
    text: String
        A text to print with optional richer format.
    warning: Bool
        If True, format as warning text (orange color).
    danger: Bool
        If True, format as danger text (red color).
        If True, overrides warning.
    format: String
        If None return plain text.
        If 'html' return HTML formatted text.
        If 'rich' return formatted text to be printed with prompt_toolkit.

    See also
    --------
    gen_tso.pandeia_io.tso_print

    Examples
    --------
    >>> import gen_tso.utils as u
    >>> text = 'WASP-80 b'
    >>> plain = u.format_text(text, danger=True)
    >>> normal = u.format_text(text, warning=False, danger=False, format='html')
    >>> html = u.format_text(text, danger=True, format='html')
    >>> rich = u.format_text(text, danger=True, format='rich')

    >>> warned = u.format_text(text, warning=True, format='html')
    >>> danger1 = u.format_text(text, danger=True, format='html')
    >>> danger2 = u.format_text(text, warning=True, danger=True, format='html')
    """
    status = 'normal'
    if danger:
        status = 'danger'
    elif warning:
        status = 'warning'

    if format is None or status=='normal':
        return text

    if format == 'html':
        text_value = f'<span class="{status}">{text}</span>'
    elif format == 'rich':
        text_value = f'<{status}>{text}</{status}>'
    return text_value


def pretty_print_target(target):
    """
    Print a target's info to HTML text.
    Must look pretty.
    """
    rplanet = as_str(target.rplanet, '.3f', '---')
    mplanet = as_str(target.mplanet, '.3f', '---')
    sma = as_str(target.sma, '.3f', '---')
    rprs = as_str(target.rprs, '.3f', '---')
    ars = as_str(target.ars, '.3f', '---')
    period = as_str(target.period, '.3f', '---')
    t_dur = as_str(target.transit_dur, '.3f', '---')
    eq_temp = as_str(target.eq_temp, '.1f', '---')

    rstar = as_str(target.rstar, '.3f', '---')
    mstar = as_str(target.mstar, '.3f', '---')
    logg = as_str(target.logg_star, '.2f', '---')
    metal = as_str(target.metal_star, '.2f', '---')
    teff = as_str(target.teff, '.1f', '---')
    ks_mag = as_str(target.ks_mag, '.2f', '---')

    status = 'confirmed' if target.is_confirmed else 'candidate'
    mplanet_label = 'M*sin(i)' if target.is_min_mass else 'mplanet'
    if len(target.aliases) > 0:
        aliases = f'aliases = {target.aliases}'
    else:
        aliases = ''

    color = '#15B01A' if target.is_jwst_planet else 'black'
    jwst_planet = f'<span style="color:{color}">{target.is_jwst_planet}</span>'
    color = '#15B01A' if target.is_jwst_host else 'black'
    jwst_host = f'<span style="color:{color}">{target.is_jwst_host}</span>'

    planet_info = ui.HTML(
        f'planet = {repr(target.planet)}<br>'
        f'is_jwst_planet = {jwst_planet}<br>'
        f'is_transiting = {target.is_transiting}<br>'
        f"status = '{status} planet'<br><br>"
        f"rplanet = {rplanet} r_earth<br>"
        f"{mplanet_label} = {mplanet} m_earth<br>"
        f"semi_major_axis = {sma} AU<br>"
        f"period = {period} d<br>"
        f"equilibrium_temp = {eq_temp} K<br>"
        f"transit_duration = {t_dur} h<br>"
        f"rplanet/rstar = {rprs}<br>"
        f"a/rstar = {ars}<br>"
    )

    star_info = ui.HTML(
        f'host = {repr(target.host)}<br>'
        f'is_jwst_host = {jwst_host}<br>'
        f'<br><br><br>'
        f"rstar = {rstar} r_sun<br>"
        f"mstar = {mstar} m_sun<br>"
        f"log_g = {logg}<br>"
        f"metallicity = {metal}<br>"
        f"effective_temp = {teff} K<br>"
        f"Ks_mag = {ks_mag}<br>"
        f"RA = {target.ra:.3f} deg<br>"
        f"dec = {target.dec:.3f} deg<br>"
    )

    return planet_info, star_info, aliases






# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import gen_tso.pandeia_io as jwst
from pandeia.engine.calc_utils import get_instrument_config
import pytest
import numpy as np


@pytest.mark.parametrize(
    'configs',
    (
        # inst,  mode, aperture, expected_rn
        ('miri', 'mrs_ts', 'ch1', 32.6),
        ('miri', 'mrs_ts', 'ch2', 32.6),
        ('miri', 'mrs_ts', 'ch3', 32.6),
        ('miri', 'mrs_ts', 'ch4', 32.6),
        ('miri', 'lrsslitless', 'imager', 32.6),
        ('miri', 'imaging_ts', 'imager', 32.6),
        ('nircam', 'lw_tsgrism', 'lw', 8.98),
        ('nircam', 'sw_tsgrism', 'dhs0spec2', 8.98),
        ('nircam', 'sw_tsgrism', 'dhs0spec4', 8.98),
        ('nircam', 'sw_tsgrism', 'dhs0spec8', 8.98),
        ('nircam', 'sw_tsgrism', 'dhs0bright', 8.98),
        ('nircam', 'lw_ts', 'lw', 8.98),
        ('nircam', 'sw_ts', 'sw', 8.98),
        ('nircam', 'sw_ts', 'wlp4', 8.98),
        ('nircam', 'sw_ts', 'wlp8__ts', 8.98),
        ('nircam', 'sw_ts', 'wlp8__tsgrism', 8.98),
        ('nirspec', 'bots',  's1600a1', 9.799),
        ('niriss', 'soss', 'soss', 11.55),
    )
)
def test_read_noise_variance(configs):
    inst, mode, aperture, expected_rn = configs
    ins_config = get_instrument_config('jwst', inst)
    # mock report
    instrument = {
        'mode': mode,
        'aperture': aperture,
    }
    configuration = {'instrument': instrument}
    input = {'configuration': configuration}
    report = {'input': input}

    read_noise = jwst.read_noise_variance(report, ins_config)
    np.testing.assert_almost_equal(read_noise, expected_rn)


def test_work_with_run_pandeia_outputs():
    import pickle

    # TSO
    tso_file = 'tso_nircam_1251.pickle'
    with open(tso_file, 'rb') as f:
        tso = pickle.load(f)
    report_tso = tso['tso']

    # TA
    tso_file = 'tso_nircam_ta.pickle'
    with open(tso_file, 'rb') as f:
        tso = pickle.load(f)
    report_ta = tso['tso']

    # Warning
    tso_file = 'tso_nircam_ta2.pickle'
    with open(tso_file, 'rb') as f:
        tso = pickle.load(f)
    report_warn = tso['tso']

    warnings = report['warnings']

    # Extracted S/N ratio: {x}
    # Max. fraction of saturation: {X}%



def test_set_scene_phoenix():
    pass


def test_set_scene_kurucz():
    pass


def test_set_scene_blackbody():
    pass


def test_set_scene_background():
    pass


def test_extract_sed_no_normalization():
    # Lets look directly at the sun:

    # 'G5V 5700K log(g)=4.5'
    sed_type = 'phoenix'
    sed_model = 'g5v'
    norm_band = '2mass,ks'
    norm_magnitude = 8.637
    scene = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    scene['spectrum']['normalization'] = dict(type="none")
    wl, flux_no = jwst.extract_sed(scene)

    # 'G5V 5770K log(g)=4.5'
    sed_type = 'k93models'
    sed_model = 'g5v'
    scene2 = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    scene2['spectrum']['normalization'] = dict(type="none")
    k_wl, k_flux_no = jwst.extract_sed(scene2)

    sed_type = 'blackbody'
    sed_model = 5770.0
    scene3 = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    scene3['spectrum']['normalization'] = dict(type="none")
    b_wl, b_flux_no = jwst.extract_sed(scene3)

    # Raw outputs
    plt.figure(0)
    plt.clf()
    plt.plot(wl, flux_no, c='orange', alpha=0.5)
    plt.plot(k_wl, k_flux_no, c='r')
    plt.plot(b_wl, b_flux_no, c='b')
    plt.xlim(0.1, 30)
    plt.xscale('log')


def test_acquisition_calculation():

    pando = jwst.PandeiaCalculation('nircam', 'target_acq')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    result = pando.perform_calculation(ngroup=3, nint=1)

    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    result = pando.perform_calculation(
        ngroup=92, nint=683,
        readout='rapid', filter='f444w',
    )
    output = result['scalar']

    # Exposure time: 0.06 s (0.00 h)
    # Max. fraction of saturation: 32.0%
    # ngroup below 80% saturation: 7
    # ngroup below 100% saturation: 9
    print(print_pandeia_report(output))


def test_report_acquisition():
    import gen_tso.pandeia_io as jwst
    pando = jwst.PandeiaCalculation('nircam', 'target_acq')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    tso = pando.perform_calculation(ngroup=3, nint=1)

    if isinstance(tso, list):
        report_in = [report['report_in']['scalar'] for report in tso]
        report_out = [report['report_out']['scalar'] for report in tso]
        reports = report_in, report_out
    else:
        reports = tso['report_in']['scalar'], tso['report_out']['scalar']

    report_in = tso['scalar']
    tso_report = print_pandeia_report(report_in)
    print(tso_report)


def test_report_single_tso():
    import gen_tso.pandeia_io as jwst
    wl = np.logspace(0, 2, 1000)
    depth = [wl, np.tile(0.03, len(wl))]

    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    tso = pando.tso_calculation(
        'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
        ngroup=90, readout='rapid', filter='f444w',
    )

    if isinstance(tso, list):
        report_in = [report['report_in']['scalar'] for report in tso]
        report_out = [report['report_out']['scalar'] for report in tso]
        reports = report_in, report_out
    else:
        reports = tso['report_in']['scalar'], tso['report_out']['scalar']

    report_in, report_out = reports
    tso_report = print_pandeia_report(report_in, report_out)
    print(tso_report)


def test_report_tso_list():
    import gen_tso.pandeia_io as jwst
    wl = np.logspace(0, 2, 1000)
    depth = [wl, np.tile(0.03, len(wl))]

    pando = jwst.PandeiaCalculation('miri', 'mrs_ts')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    aperture = ['ch1', 'ch2', 'ch3', 'ch4']
    tso = pando.tso_calculation(
        'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
        ngroup=250, aperture=aperture,
    )

    pixel_rate, full_well = jwst.saturation_level(tso)

    if isinstance(tso, list):
        report_in = [report['report_in']['scalar'] for report in tso]
        report_out = [report['report_out']['scalar'] for report in tso]
        reports = report_in, report_out
    else:
        reports = tso['report_in']['scalar'], tso['report_out']['scalar']

    report_in, report_out = reports
    tso_report = print_pandeia_report(report_in, report_out)
    print(tso_report)


def test_tso_calculation():
    import pandeia_interface as jwst
    from pandeia_interface import set_depth_scene

    # With NIRCam
    instrument = 'nircam'
    mode = 'lw_tsgrism'
    self = pando = jwst.PandeiaCalculation(instrument, mode)

    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.637,
    )
    disperser='grismr'
    filter='f444w'
    readout='rapid'
    subarray='subgrism64'
    ngroup = 127

    # With NIRSpec
    instrument = 'nirspec'
    mode = 'bots'
    self = pando = jwst.PandeiaCalculation(instrument, mode)

    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=18.637,
    )
    disperser='g395h'
    filter='f290lp'
    readout='nrsrapid'
    subarray='sub2048'
    ngroup = 16


    transit_dur = 2.753
    obs_dur = 6.0
    obs_type = 'transit'

    depth_model = np.loadtxt(
        '../planet_spectra/WASP80b_transit.dat', unpack=True)

    figure(0)
    clf()
    plot(wl, var_in, c='b')
    plot(wl, in_transit_shot_var, c='orange')
    plot(wl, in_transit_shot_var*5/6, c='0.5')
    plot(wl, in_transit_flux, c='xkcd:green')
    plot(wl, in_transit_background_var, c='black')
    axhline(in_transit_rn_var, c='red')

    sed_type = 'k93models'
    sed_model = 'k7v'
    norm_band = '2mass,ks'
    norm_magnitude = 8.637
    scene = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    wl2, kur = jwst.extract_sed(scene)

    plt.clf()
    plt.plot(wl1, phoenix)
    plt.plot(wl2, kur)
    plt.xlim(0.5, 50)
    plt.xscale('log')

    plt.clf()
    plt.plot(wl1[1:], wl1[1:]/np.ediff1d(wl1))
    plt.plot(wl2[1:], wl2[1:]/np.ediff1d(wl2))
    plt.xlim(0.5, 15)
    plt.yscale('log')


    # # Check flux rates are OK
    # plt.figure(1)
    # plt.clf()
    # plt.plot(wl, flux_in/dt_in, c='orange', label='in transit')
    # plt.plot(wl, flux_out/dt_out, c='b', alpha=0.7, label='out transit')
    # plt.legend(loc='best')

    # # Check output depth spectrum matches input depth spectrum
    # wl_depth, depth = depth_model
    # lw = 1.0
    # plt.figure(4)
    # plt.clf()
    # plt.plot()
    # plt.plot(wl_depth, 100*depth, c='orange', lw=lw, label='model depth')
    # plt.plot(wl, 100*obs_depth, c='b', alpha=0.7, lw=lw, label='obs depth')
    # plt.legend(loc='best')
    # plt.xlim(2.7, 5.2)
    # plt.ylim(2.8, 3.2)

    # # Check flux rates are OK
    # plt.figure(4)
    # plt.clf()
    # plt.plot(wl, flux_in/dt_in, c='orange', label='in transit')
    # plt.plot(wl, var_in/dt_in, c='xkcd:green', label='in transit')
    # plt.legend(loc='best')

    #plt.plot(wl, in_transit_shot_var, c='orange')
    #plt.plot(wl, in_transit_shot_var*5/6, c='0.5')
    #plt.plot(wl, in_transit_flux, c='xkcd:green')
    #plt.plot(wl, in_transit_background_var, c='black')
    #plt.axhline(in_transit_rn_var, c='red')

    #'e_rate_out':photon_out_bin/to,
    #'e_rate_in':photon_in_bin/ti,
    #'error_no_floor':error_spec,
    #'rn[out,in]': result['rn[out,in]'],
    #'bkg[out,in]': result['bkg[out,in]']




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'catalogs',
    'pandeia_io',
    'plotly_io',
    'custom_shiny',
]

from . import catalogs
from . import pandeia_io
from . import plotly_io
from . import custom_shiny
from .version import __version__


# Clean up top-level namespace--delete everything that isn't in __all__
# or is a magic attribute, and that isn't a submodule of this package
for varname in dir():
    if not ((varname.startswith('__') and varname.endswith('__')) or
            varname in __all__):
        del locals()[varname]
del(varname)




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import textwrap

import numpy as np
import gen_tso.pandeia_io as jwst
from gen_tso.app_utils import (
    planet_model_name,
    parse_instrument,
    parse_obs,
    parse_sed,
)


warning_template = (
    "# NOTE! Need to set the path to the FILE.\n    "
    "# (browsers don't expose paths of uploaded files for security reasons)\n    "
)


def parse_depth_source(input, spectra, teq_text=None, rprs_text=None, depth_text=None):
    """
    Parse transit/eclipse model name based on current state.
    Calculate or extract model.
    """
    model_type = input.planet_model_type.get()
    depth_label = planet_model_name(input)
    obs_geometry = input.obs_geometry.get()

    transit_depth_script = ""

    if model_type == 'Input':
        if depth_label is None:
            alter = 'flat transit' if obs_geometry=='transit' else 'blackbody'
            return  f"""
    # NOTE!  Need to upload a {obs_geometry}-depth spectrum to simulate a TSO
    # or alternatively select a {alter} model
"""
        model = spectra[obs_geometry][depth_label]
        filename = model['filename']
        units = model['units']

        warning = ""
        if 'unknown_' in filename:
            filename = filename.replace('unknown_', '')
            warning = warning_template.replace('FILE', f"{obs_geometry} 'depth_file'")

        transit_depth_script = f"""
    # The planet's {obs_geometry} spectrum:
    {warning}obs_type = {repr(obs_geometry)}
    units = {repr(units)}
    depth_file = {repr(filename)}
    label, wl, depth = u.read_spectrum_file(depth_file, units)
    depth_model = [wl, depth]"""

        return transit_depth_script

    if model_type == 'Flat':
        if depth_text is None:
            depth = input.transit_depth.get() * 0.01
            depth_text = f'{depth:.5e}'
        transit_depth_script = f"""
    # The planet's {obs_geometry} spectrum:
    obs_type = {repr(obs_geometry)}
    wl = ps.constant_resolution_spectrum(0.1, 50.0, resolution=300)
    nwave = len(wl)
    depth = np.tile({depth_text}, nwave)
    depth_model = [wl, depth]"""

        return transit_depth_script

    elif model_type == 'Blackbody':
        if rprs_text is None:
            rprs = np.sqrt(input.eclipse_depth.get() * 0.01)
            rprs_text = f'{rprs:.5f}'
        if teq_text is None:
            teq_text = f'{input.teq_planet.get():.1f}'
        transit_depth_script = f"""
    # The planet's {obs_geometry} spectrum:
    obs_type = {repr(obs_geometry)}
    t_planet = {teq_text}
    rprs = {rprs_text}
    wl, depth = jwst.blackbody_eclipse_depth(t_planet, rprs, sed_type, sed_model)
    depth_model = [wl, depth]"""

        return transit_depth_script



def export_script_fixed_values(
        input, spectra, saturation_fraction,
        acquisition_targets, acq_target_list,
    ):
    """Translate gen_tso's current app state to a python script"""
    config = parse_instrument(
        input, 'instrument', 'mode', 'aperture', 'disperser', 'filter',
        'subarray', 'readout', 'order', 'ngroup', 'nint',
        'pairing', 'pupil', 'detector',
    )
    inst, mode, aperture, disperser, filter, subarray, readout = config[0:7]
    order, ngroup, nint, pairing, pupil, detector = config[7:]

    name = input.target.get()
    target_focus = input.target_focus.get()
    if target_focus == 'acquisition':
        selected = acquisition_targets.cell_selection()['rows'][0]
        target_list = acq_target_list.get()
        target_acq_mag = np.round(target_list[1][selected], 3)
        name = target_list[0][selected]
    elif target_focus == 'science':
        target_acq_mag = None

    sed_type, sed_model, norm_band, norm_mag, sed_label = parse_sed(
        input, spectra, target_acq_mag=target_acq_mag,
    )
    sed_warning = ""
    if sed_type == 'input':
        sed_units = sed_model['units']
        sed_file = sed_model['filename']
        if 'unknown_' in sed_file:
            sed_file = sed_file.replace('unknown_', '')
            sed_warning = warning_template.replace('FILE', "SED's 'sed_file'")

        sed_script = f"""
    sed_units = {repr(sed_units)}
    sed_file = {repr(sed_file)}
    label, sed_wl, flux = u.read_spectrum_file(sed_file, sed_units)
    sed_model = {{'wl': sed_wl, 'flux': flux}}
    """.strip()
    elif sed_type == 'blackbody':
        t_eff = input.t_eff.get()
        sed_script = f"sed_model = {t_eff}"
    else:
        sed_script = f"sed_model = {repr(sed_model)}"

    # Write the script
    script = f"""
    # The Pandeia instrumental configuration:
    instrument = {repr(inst)}
    mode = {repr(mode)}
    pando = jwst.PandeiaCalculation(instrument, mode)

    disperser = {repr(disperser)}
    filter = {repr(filter)}
    subarray = {repr(subarray)}
    readout = {repr(readout)}
    aperture = {repr(aperture)}
    order = {repr(order)}
    ngroup = {ngroup}
    nint = {repr(nint)}

    # The target ({name}):
    {sed_warning}sed_type = {repr(sed_type)}
    {sed_script}
    norm_band = {repr(norm_band)}
    norm_mag = {norm_mag}
    pando.set_scene(sed_type, sed_model, norm_band, norm_mag)\
"""

    if mode == 'target_acq':
        script += """\n
    # Run target acquisition
    tso = pando.perform_calculation(
        ngroup, nint, disperser, filter, subarray, readout,
        aperture, order,
    )\
"""
    else:
        # Transit depth
        transit_depth_script = parse_depth_source(input, spectra)
        transit_dur = float(input.t_dur.get())
        script += f"""
    {transit_depth_script}

    # in-transit and total observation duration times (hours):
    transit_dur = {transit_dur}
    exp_time = jwst.exposure_time(instrument, subarray, readout, ngroup, nint)
    obs_dur = exp_time / 3600.0

    # Run TSO simulation:
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout, aperture, order,
    )\
"""

    imports = """\
    import gen_tso.pandeia_io as jwst
    import gen_tso.utils as u
"""

    if 'np.' in script:
        imports += "    import numpy as np\n"
    if 'ps.' in script:
        imports += "    import pyratbay.spectrum as ps\n"
    script = imports + script
    return textwrap.dedent(script)


def export_script_calculated_values(
        input, spectra, saturation_fraction,
        acquisition_targets, acq_target_list, catalog,
    ):
    """
    Translate gen_tso's current app state to a python script
    When values can be computed from variables rather than fixed
    numbers, show that script.
    """
    config = parse_instrument(
        input, 'instrument', 'mode', 'aperture', 'disperser', 'filter',
        'subarray', 'readout', 'order', 'ngroup', 'nint',
        'pairing', 'pupil', 'detector',
    )
    inst, mode, aperture, disperser, filter, subarray, readout = config[0:7]
    order, ngroup, nint, pairing, pupil, detector = config[7:]

    name = input.target.get()
    target = catalog.get_target(name)
    target_focus = input.target_focus.get()
    if target_focus == 'acquisition':
        selected = acquisition_targets.cell_selection()['rows'][0]
        target_list = acq_target_list.get()
        target_acq_mag = np.round(target_list[1][selected], 3)
        name = target_list[0][selected]
    elif target_focus == 'science':
        target_acq_mag = None

    req_saturation = saturation_fraction.get()
    transit_dur = float(input.t_dur.get())
    planet_model_type, depth_label, rprs_sq, teq_planet = parse_obs(input)

    sed_type, sed_model, norm_band, norm_mag, sed_label = parse_sed(
        input, spectra, target_acq_mag=target_acq_mag,
    )
    sed_warning = ""
    t_eff = float(input.t_eff.get())
    log_g = float(input.log_g.get())
    teff_text = 'target.teff' if t_eff == target.teff else f'{t_eff}'
    logg_text = 'target.logg_star' if log_g == target.logg_star else f'{log_g}'

    # Magnitude
    is_target_kmag = (
        target_focus == 'science' and
        norm_band == '2mass,ks' and
        norm_mag == target.ks_mag
    )
    mag_text = 'target.ks_mag' if is_target_kmag else f'{norm_mag:.4f}'

    # SED
    if sed_type == 'input':
        sed_units = sed_model['units']
        sed_file = sed_model['filename']
        if 'unknown_' in sed_file:
            sed_file = sed_file.replace('unknown_', '')
            sed_warning = warning_template.replace('FILE', "SED's 'sed_file'")

        sed_script = f"""
    sed_units = {repr(sed_units)}
    sed_file = {repr(sed_file)}
    label, sed_wl, flux = u.read_spectrum_file(sed_file, sed_units)
    sed_model = {{'wl': sed_wl, 'flux': flux}}
    """.strip()
    elif sed_type == 'blackbody':
        sed_script = f"sed_model = {teff_text}"
    else:
        calc_sed_model = jwst.find_closest_sed(t_eff, log_g, sed_type)
        if target_focus == 'acquisition' or calc_sed_model != sed_model:
            sed_script = f"sed_model = {repr(sed_model)}"
        else:
            sed_script = (
                f"t_eff = {teff_text}\n    "
                f"logg_star = {logg_text}\n    "
                "sed_model = jwst.find_closest_sed(t_eff, logg_star, sed_type)"
            )

    sed_script = f"""
    {sed_warning}sed_type = {repr(sed_type)}
    {sed_script}
    norm_band = {repr(norm_band)}
    norm_mag = {mag_text}
    pando.set_scene(sed_type, sed_model, norm_band, norm_mag)
"""

    # ngroup
    pando = jwst.PandeiaCalculation(inst, mode)
    pando.set_config(disperser, filter, subarray, readout, aperture, order)
    pando.set_scene(sed_type, sed_model, norm_band, norm_mag)
    ngroup_script = (
        '# Set ngroup below requested saturation fraction '
        f'(ngroup = {ngroup}):\n    '
        f'ngroup = pando.saturation_fraction(fraction={req_saturation:.1f})'
    )
    if ngroup != pando.saturation_fraction(fraction=req_saturation):
        ngroup_script = (
            '# Number of groups per integration:\n    '
            f'ngroup = {ngroup}'
        )

    # obs_duration
    obs_dur = float(input.obs_dur.get())
    transit_dur = float(input.t_dur.get())
    is_target_tdur = np.abs(transit_dur - target.transit_dur) < 0.01
    tdur_text = 'target.transit_dur' if is_target_tdur else f'{transit_dur}'

    t_start = 1.0
    t_settling = input.settling_time.get()
    t_base = input.baseline_time.get()
    min_baseline = input.min_baseline_time.get()
    t_baseline = np.max([t_base*transit_dur, min_baseline])
    total_duration = t_start + t_settling + transit_dur + 2*t_baseline
    if np.abs(total_duration - obs_dur) < 0.01:
        time_script = f"""
    # Estimate in-transit and total duration of observation:
    transit_dur = {tdur_text}
    t_start = 1.0
    t_settling = {t_settling}
    t_base = np.max([{t_base}*transit_dur, {min_baseline}])
    total_duration = t_start + t_settling + transit_dur + 2*t_base
"""
    else:
        time_script = (
             '\n    # in-transit and total duration of observation:'
            f'\n    transit_dur = {tdur_text}'
            f'\n    total_duration = {obs_dur}'
        )

    # nint
    match_integs = input.integs_switch.get()
    calculated_nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_dur,
    )
    if match_integs and nint == calculated_nint:
        nint_script = (
            f'{time_script}\n    '
            f'# Set nint to match the observation duration (nint = {nint}):\n'
            '    nint, exp_time = jwst.bin_search_exposure_time(\n'
            '        instrument, subarray, readout, ngroup, total_duration,\n'
            '    )'
        )
    else:
        nint_script = f"\n    # Number of integrations:\n    nint = {nint}"

    # Instrument setup
    inst_script = f"""
    # The Pandeia instrumental configuration:
    instrument = {repr(inst)}
    mode = {repr(mode)}
    pando = jwst.PandeiaCalculation(instrument, mode)

    disperser = {repr(disperser)}
    filter = {repr(filter)}
    subarray = {repr(subarray)}
    readout = {repr(readout)}
    aperture = {repr(aperture)}
    order = {repr(order)}
    pando.set_config(disperser, filter, subarray, readout, aperture, order)

    # The target ({name}):\
"""

    teq = input.teq_planet.get()
    teq_text = 'target.eq_temp' if np.abs(teq-target.eq_temp)<1.0 else f'{teq}'

    rprs = np.sqrt(input.eclipse_depth.get() * 0.01)
    is_target_rprs = np.abs(rprs - target.rprs) < 0.001
    rprs_text = 'target.rprs' if is_target_rprs else f'{rprs:.5f}'

    depth = input.transit_depth.get() * 0.01
    is_target_depth = np.abs(np.sqrt(depth) - target.rprs) < 0.001
    depth_text = 'target.rprs**2' if is_target_depth else f'{depth:.5e}'

    if mode == 'target_acq':
        sed_script += f"""
    # Timings
    ngroup = {ngroup}
    nint = 1

    # Run target acquisition
    tso = pando.perform_calculation(ngroup, nint)\
"""
    else:
        # Transit depth
        transit_depth_script = parse_depth_source(input, spectra, teq_text, rprs_text, depth_text)
        sed_script += f"""
    {ngroup_script}
    {nint_script}
    # Observation duration times (hours):
    exp_time = jwst.exposure_time(instrument, subarray, readout, ngroup, nint)
    obs_dur = exp_time / 3600.0
    {transit_depth_script}

    # Run TSO simulation:
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model, ngroup,
    )\
"""

    # Now put everything together:
    target_script = f"""
    catalog = cat.Catalog()
    target = catalog.get_target({repr(name)})
"""
    if 'target.' in sed_script:
        script = inst_script + target_script + sed_script
    else:
        script = inst_script + sed_script

    imports = """\
    import gen_tso.pandeia_io as jwst
    import gen_tso.utils as u
"""
    if 'cat.' in script:
        imports += "    import gen_tso.catalogs as cat\n"
    if 'ps.' in script:
        imports += "    import pyratbay.spectrum as ps\n"
    if 'np.' in script:
        imports += "    import numpy as np\n"
    script = imports + script

    return textwrap.dedent(script)



# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'Target',
]

import numpy as np
import pyratbay.constants as pc
import pyratbay.atmosphere as pa

from . import utils as u


class Target():
    """
    A handy exoplanet target object.
    """
    def __init__(
        self, entry=None,
        host=None, mstar=np.nan, rstar=np.nan, teff=np.nan, logg_star=np.nan,
        metal_star=np.nan, ks_mag=np.nan, ra=np.nan, dec=np.nan,
        planet=None, mplanet=np.nan, rplanet=np.nan, period=np.nan, sma=np.nan,
        transit_dur=np.nan, ars=np.nan, rprs=np.nan, eq_temp=np.nan,
        is_confirmed=np.nan, is_min_mass=False, aliases=[],
    ):
        # Turn on if planet mass is RV's minimum mass: M*sin(i)
        self.is_min_mass = False
        if entry is None:
            if host is None or planet is None:
                raise ValueError('Must at least specify planet and host names')
            self.host = host
            self.mstar = mstar
            self.rstar = rstar
            self.teff = teff
            self.logg_star = logg_star
            self.metal_star = metal_star
            self.ks_mag = ks_mag
            self.ra = ra
            self.dec = dec

            self.planet = planet
            self.mplanet = mplanet
            self.rplanet = rplanet
            self.sma = sma
            self.period = period
            self.ars = ars
            self.rprs = rprs
            self.eq_temp = eq_temp
            self.transit_dur = transit_dur
            self.is_min_mass = is_min_mass

        else:
            self.host = entry['hostname']
            self.mstar = entry['st_mass']
            self.rstar = entry['st_rad']
            self.teff = entry['st_teff']
            self.logg_star = entry['st_logg']
            self.metal_star = entry['st_met']
            self.ks_mag = entry['sy_kmag']
            self.ra = entry['ra']
            self.dec = entry['dec']

            self.planet = entry['pl_name']
            self.mplanet = entry['pl_masse']
            if np.isnan(entry['pl_masse']) and np.isfinite(entry['pl_msinie']):
                self.mplanet = entry['pl_msinie']
                self.is_min_mass = True
            self.rplanet = entry['pl_rade']
            self.sma = entry['pl_orbsmax']
            self.period = entry['pl_orbper']
            self.ars = entry['pl_ratdor']
            self.rprs = entry['pl_ratror']
            self.transit_dur = entry['pl_trandur']
            self.eq_temp = entry['pl_eqt']

        self.is_transiting = np.isfinite(self.transit_dur)
        self.is_confirmed = is_confirmed
        self.aliases = aliases
        self._complete_values()

    def _complete_values(self):
        # Rank groups
        ars = (
            np.isnan(self.sma) +
            np.isnan(self.ars) +
            np.isnan(self.rstar)
        )
        rprs = (
            np.isnan(self.rplanet) +
            np.isnan(self.rprs) +
            np.isnan(self.rstar)
        )
        aperiod = (
            np.isnan(self.period) +
            np.isnan(self.sma) +
            np.isnan(self.mstar)
        )
        solve_order = np.argsort([ars, rprs, aperiod])
        for i in solve_order:
            if i == 0:
                solution = solve_a_rs(self.sma, self.rstar, self.ars)
                self.sma, self.rstar, self.ars = solution
            if i == 1:
                solution = solve_rprs(self.rplanet, self.rstar, self.rprs)
                self.rplanet, self.rstar, self.rprs = solution
            if i == 2:
                solution = solve_sma_period(self.period, self.sma, self.mstar)
                self.period, self.sma, self.mstar = solution
        self.equilibrium_temp()


    def equilibrium_temp(self, A=0.0, recirculation=1.0, force_update=False):
        """
        Parameters
        ----------
        A: Scalar
            Planetary bond albedo.
        recirculation: Scalar
            Day--night energy recirculation factor:
            recirculation = 0.5  no redistribution (total dayside reemission)
            recirculation = 1.0  good redistribution (4pi reemission)
        """
        eq_temp, _ = pa.equilibrium_temp(
            self.teff,
            self.rstar*pc.rsun,
            self.sma*pc.au,
            A=A, f=recirculation,
        )
        update = (
            np.isfinite(eq_temp) and
            (np.isnan(self.eq_temp) or force_update)
        )
        if update:
            self.eq_temp = eq_temp
        return eq_temp

    def copy_star(self, star):
        props = [
            'rstar',
            'mstar',
            'teff',
            'metal_star',
            'logg_star',
            'ks_mag',
        ]
        update_rplanet = (
            np.isfinite(self.rprs) and
            np.isfinite(self.rstar) and
            np.isfinite(star.rstar) and
            self.rstar != star.rstar
        )
        for prop in props:
            if np.isfinite(getattr(star, prop)):
                setattr(self, prop, getattr(star, prop))
        if update_rplanet:
            self.rplanet = self.rprs * self.rstar*pc.rsun / pc.rearth
        # Now complement all other values
        self._complete_values()

    def machine_readable_text(self):
        is_jwst_planet = is_jwst_host = ''
        if hasattr(self, 'is_jwst_planet'):
            is_jwst_planet = f'is_jwst_planet = {self.is_jwst_planet}\n'
            is_jwst_host = f'is_jwst_host = {self.is_jwst_host}\n'
        status = 'confirmed planet' if self.is_confirmed else 'candidate'
        mplanet_label = 'm_sini' if self.is_min_mass else 'mplanet'

        rstar = u.as_str(self.rstar, '.3f', 'np.nan')
        mstar = u.as_str(self.mstar, '.3f', 'np.nan')
        teff = u.as_str(self.teff, '.1f', 'np.nan')
        log_g = u.as_str(self.logg_star, '.2f', 'np.nan')
        metal = u.as_str(self.metal_star, '.2f', 'np.nan')
        ks_mag = u.as_str(self.ks_mag, '.2f', 'np.nan')
        ra = u.as_str(self.ra, '.4f', 'np.nan')
        dec = u.as_str(self.dec, '.4f', 'np.nan')

        rplanet = u.as_str(self.rplanet, '.3f', 'np.nan')
        mplanet = u.as_str(self.mplanet, '.3f', 'np.nan')
        rprs = u.as_str(self.rprs, '.3f', 'np.nan')
        a_rstar = u.as_str(self.ars, '.3f', 'np.nan')
        sma = u.as_str(self.sma, '.3f', 'np.nan')
        period = u.as_str(self.period, '.3f', 'np.nan')
        transit_dur = u.as_str(self.transit_dur, '.3f', 'np.nan')
        eq_temp = u.as_str(self.eq_temp, '.1f', 'np.nan')

        report = (
            f"host = {repr(self.host)}\n"
            f"planet = {repr(self.planet)}\n"
            f'aliases = {self.aliases}\n\n'

            f"rstar = {rstar}  # r_sun\n"
            f"mstar = {mstar}  # m_sun\n"
            f"teff = {teff} # K\n"
            f"log_g = {log_g}\n"
            f"metallicity = {metal}\n"
            f"ks_mag = {ks_mag}\n"
            f"ra = {ra}  # deg\n"
            f"dec = {dec}  # deg\n\n"

            f"rplanet = {rplanet}  # r_earth\n"
            f"{mplanet_label} = {mplanet}  # m_earth\n"
            f"transit_dur = {transit_dur}  # h\n"
            f"sma = {sma}  # AU\n"
            f"period = {period}  # d\n"
            f"eq_temp = {eq_temp}  # K\n"
            f"rprs = {rprs}\n"
            f"a_rstar = {a_rstar}\n"

            f'\nis_transiting = {self.is_transiting}\n'
            f'{is_jwst_planet}'
            f'{is_jwst_host}'
            f'status = {repr(status)}\n'
        )
        return report


    def __str__(self):
        rstar = u.as_str(self.rstar, '.3f', '---')
        mstar = u.as_str(self.mstar, '.3f', '---')
        teff = u.as_str(self.teff, '.1f', '---')
        logg = u.as_str(self.logg_star, '.2f', '---')
        metal = u.as_str(self.metal_star, '.2f', '---')
        ks_mag = u.as_str(self.ks_mag, '.2f', '---')
        ra = u.as_str(self.ra, '.3f', '---')
        dec = u.as_str(self.dec, '.3f', '---')

        rplanet = u.as_str(self.rplanet, '.3f', '---')
        mplanet = u.as_str(self.mplanet, '.3f', '---')
        rprs = u.as_str(self.rprs, '.3f', '---')
        ars = u.as_str(self.ars, '.3f', '---')
        sma = u.as_str(self.sma, '.3f', '---')
        period = u.as_str(self.period, '.3f', '---')
        t14 = u.as_str(self.transit_dur, '.3f', '---')
        eq_temp = u.as_str(self.eq_temp, '.1f', '---')

        mplanet_label = 'M*sin(i)' if self.is_min_mass else 'mplanet'
        report = (
            f"planet = {self.planet}\n"
            f"host = {self.host}\n\n"
            f"rstar = {rstar} r_sun\n"
            f"mstar = {mstar} m_sun\n"
            f"teff = {teff} K\n"
            f"log_g = {logg}\n"
            f"metallicity = {metal}\n"
            f"Ks_mag = {ks_mag}\n"
            f"RA = {ra} deg\n"
            f"dec = {dec} deg\n\n"

            f"rplanet = {rplanet} r_earth\n"
            f"{mplanet_label} = {mplanet} m_earth\n"
            f"transit_dur = {t14} h\n"
            f"sma = {sma} AU\n"
            f"period = {period} d\n"
            f"eq_temp = {eq_temp} K\n"
            f"rplanet/rstar = {rprs}\n"
            f"a/rstar = {ars}\n"
        )
        report += f'\nis transiting = {self.is_transiting}\n'
        if hasattr(self, 'is_jwst'):
            report += f'is JWST host = {self.is_jwst}\n'
        if np.isfinite(self.is_confirmed):
            status = 'confirmed' if self.is_confirmed else 'candidate'
            report += f'status = {status} planet\n'
        if len(self.aliases) > 0:
            report += f'aliases: {self.aliases}'
        return report


def format_nea_entry(entry):
    # Have TOI entries the same keys as PS entries:
    if 'toi' in entry.keys():
        entry['hostname'] = f"TOI-{entry['toipfx']}"
        entry['st_met'] = np.nan
        entry['sy_kmag'] = np.nan
        logg = entry['st_logg'] if entry['st_logg'] is not None else np.nan
        rstar = entry['st_rad'] if entry['st_rad'] is not None else np.nan
        entry['st_mass'] = 10**logg * (rstar*pc.rsun)**2 / pc.G / pc.msun

        entry['pl_name'] = f"TOI-{entry['toi']}"
        entry['pl_masse'] = np.nan
        entry['pl_orbsmax'] = np.nan
        entry['pl_ratdor'] = np.nan
        entry['pl_ratror'] = np.sqrt(entry.pop('pl_trandep')*pc.ppm)
        entry['pl_trandur'] = entry.pop('pl_trandurh')
    else:
        entry['pl_eqt'] = np.nan

    # Patch
    if entry['st_mass'] == 0.0:
        entry['st_mass'] = np.nan

    # Replace None with np.nan
    for key in entry.keys():
        if entry[key] is None:
            entry[key] = np.nan
    return entry


def solve_sma_period(period, sma, mstar):
    """
    Solve period-sma-mstar system of equations.

    Parameters
    ----------
    period: Float
        Orbital period (days).
    sma: Float
        Orbital semi-major axis (AU).
    mstar: Float
        Stellar mass (m_sun).
    """
    missing = (
        np.isnan(period) +
        np.isnan(sma) +
        np.isnan(mstar)
    )
    # Know everything or not enough:
    if missing != 1:
        return period, sma, mstar

    two_pi_G = 2.0*np.pi / np.sqrt(pc.G)
    if np.isnan(mstar):
        mstar = (sma*pc.au)**3.0 / (period*pc.day/two_pi_G)**2.0 / pc.msun
    elif np.isnan(period):
        period = np.sqrt((sma*pc.au)**3.0 / (mstar*pc.msun)) * two_pi_G / pc.day
    elif np.isnan(sma):
        sma = ((period*pc.day/two_pi_G)**2.0 * (mstar*pc.msun))**(1/3) / pc.au

    return period, sma, mstar


def solve_rprs(rplanet, rstar, rprs):
    """
    Solve planet radius -- stellar radius system of equations.

    Parameters
    ----------
    rplanet: Float
        Planet radius (r_earth).
    rstar: Float
        Stellar radius (r_sun).
    rprs: Float
        Planet--star radius ratio.
    """
    missing = (
        np.isnan(rplanet) +
        np.isnan(rstar) +
        np.isnan(rprs)
    )
    # Know everything or not enough:
    if missing != 1:
        return rplanet, rstar, rprs

    if np.isnan(rplanet):
        rplanet = rprs * (rstar*pc.rsun) / pc.rearth
    elif np.isnan(rstar):
        rstar = rplanet*pc.rearth / rprs / pc.rsun
    elif np.isnan(rprs):
        rprs = rplanet*pc.rearth / (rstar*pc.rsun)

    return rplanet, rstar, rprs


def solve_a_rs(sma, rstar, ars):
    """
    Solve semi-major axis -- stellar radius system of equations.

    Parameters
    ----------
    sma: Float
        Orbital semi-major axis (AU).
    rstar: Float
        Stellar radius (r_sun).
    ars: Float
        sma--rstar ratio.
    """
    missing = (
        np.isnan(sma) +
        np.isnan(ars) +
        np.isnan(rstar)
    )
    # Know everything or not enough:
    if missing != 1:
       return sma, rstar, ars

    if np.isnan(sma):
        sma = ars * rstar*pc.rsun / pc.au
    elif np.isnan(rstar):
        rstar = sma*pc.au / ars / pc.rsun
    elif np.isnan(ars):
        ars = sma*pc.au / (rstar*pc.rsun)

    return sma, rstar, ars


def solve_host(targets, n_dups):
    star = targets[0]
    # Identify conflicts
    masses = [target.mstar for target in targets]
    u_masses = np.unique([mass for mass in masses if np.isfinite(mass)])

    radii = [target.rstar for target in targets]
    u_radii = np.unique([radius for radius in radii if np.isfinite(radius)])

    teff = [target.teff for target in targets]
    u_teff = np.unique([temp for temp in teff if np.isfinite(temp)])

    log_g = [target.logg_star for target in targets]
    u_logg = np.unique([grav for grav in log_g if np.isfinite(grav)])

    metal = [target.metal_star for target in targets]
    u_metal = np.unique([met for met in metal if np.isfinite(met)])

    conflicts = (
        len(u_masses) > 1,
        len(u_radii) > 1,
        len(u_teff) > 1,
        len(u_logg) > 1,
        len(u_metal) > 1,
    )
    # Solve conflicts
    if np.any(conflicts):
        props = [
            'rstar',
            'mstar',
            'teff',
            'metal_star',
            'logg_star',
        ]
        completeness = [np.sum(missing_mask(target)) for target in targets]
        # Sort by completeness then by duplicity
        rank = np.lexsort((-np.array(n_dups), completeness))
        # Set a unique set of stellar properties, complement if needed
        star = targets[rank[0]]
        for i in rank[1:]:
            target = targets[i]
            for prop in props:
                update = (
                    np.isnan(getattr(star, prop)) and
                    np.isfinite(getattr(target, prop))
                )
                if update:
                    setattr(star, prop, getattr(target, prop))
    return star


def missing_mask(target):
    missing = np.array([
        np.isnan(target.rstar),
        np.isnan(target.mstar),
        np.isnan(target.teff),
        np.isnan(target.logg_star),
        np.isnan(target.metal_star),

        np.isnan(target.rplanet),
        np.isnan(target.mplanet),
        np.isnan(target.transit_dur),
        np.isnan(target.sma),
        np.isnan(target.period),
        np.isnan(target.ars),
        np.isnan(target.rprs),
    ])
    return missing


def rank_planets(target, alt_targets):
    rank = np.zeros(len(alt_targets))

    props = [
        'rprs',
        'ars',
        'rplanet',
        'mplanet',
        'rstar',
        'mstar',
        'transit_dur',
        'period',
        'sma',
        'teff',
        'metal_star',
        'logg_star',
    ]
    t = 0
    for t in range(len(alt_targets)):
        missing = missing_mask(target)
        # rank by fill as many missing values as possible
        for i,alt in enumerate(alt_targets):
            alt_miss = missing_mask(alt)
            rank[i] = np.sum(~alt_miss & missing)
        if np.all(rank==0):
            break

        # loop by rank
        i_rank = np.argsort(-rank)[0]
        alt = alt_targets[i_rank]
        # update props
        for prop in props:
            update = (
                np.isnan(getattr(target, prop)) and
                np.isfinite(getattr(alt, prop))
            )
            if update:
                setattr(target, prop, getattr(alt, prop))
                target._complete_values()
    return t





# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import numpy as np
import gen_tso.pandeia_io as jwst


# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Spectroscopy
def test_perform_calculation_miri_lrs():
    pando = jwst.PandeiaCalculation('miri', 'lrsslitless')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    ngroup = 34
    nint = 1379

    report = pando.perform_calculation(
        ngroup, nint,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    # ETC background 2572 e/s
    np.testing.assert_allclose(rep['sn'], 2216.8585)
    np.testing.assert_allclose(rep['extracted_flux'], 2789.8679)
    np.testing.assert_allclose(rep['extracted_noise'], 1.258478, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 28412.383)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.793349, rtol=1e-6)


def test_perform_calculation_miri_mrs():
    pando = jwst.PandeiaCalculation('miri', 'mrs_ts')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    ngroup = 50
    nint = 450

    report = pando.perform_calculation(
        ngroup, nint,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 2133.817392)
    np.testing.assert_allclose(rep['extracted_flux'], 256.712868)
    np.testing.assert_allclose(rep['extracted_noise'], 0.120307, rtol=1e-5)
    np.testing.assert_allclose(rep['brightest_pixel'], 172.1364)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.123334, rtol=1e-6)


def test_perform_calculation_nirspec_bots():
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    pando.calc['background'] = 'ecliptic'

    ngroup = 14
    nint = 550

    report = pando.perform_calculation(
        ngroup, nint,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 5909.431001)
    np.testing.assert_allclose(rep['extracted_flux'], 6184.164104)
    np.testing.assert_allclose(rep['extracted_noise'], 1.046491, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 3975.6643)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.77238, rtol=1e-6)


def test_perform_calculation_niriss_soss():
    pando = jwst.PandeiaCalculation('niriss', 'soss')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    subarray = 'substrip96'
    ngroup = 8
    nint = 350

    report = pando.perform_calculation(
        ngroup, nint, subarray=subarray,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 11269.248655)
    np.testing.assert_allclose(rep['extracted_flux'], 28856.874712)
    np.testing.assert_allclose(rep['extracted_noise'], 2.560674, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 3191.468469)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.785101, rtol=1e-6)


def test_perform_calculation_nircam_lw_tsgrism():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    pando.calc['background'] = 'ecliptic'

    filter = 'f322w2'
    readout = 'rapid'
    subarray = 'subgrism64'
    ngroup = 50
    nint = 442

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray, readout=readout,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 5663.236)
    np.testing.assert_allclose(rep['extracted_flux'], 5197.943)
    np.testing.assert_allclose(rep['extracted_noise'], 0.91784, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 2370.5251)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.694858, rtol=1e-6)


def test_perform_calculation_nircam_sw_tsgrism():
    pando = jwst.PandeiaCalculation('nircam', 'sw_tsgrism')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    aperture = 'dhs0spec8'
    filter = 'f150w'
    readout = 'rapid'
    subarray = 'sub260s4_8-spectra'
    ngroup = 80
    nint = 90

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray, readout=readout,
        aperture=aperture,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 4108.861511)
    np.testing.assert_allclose(rep['extracted_flux'], 2326.930432)
    np.testing.assert_allclose(rep['extracted_noise'], 0.56632, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 485.7249)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.718261, rtol=1e-6)



# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Photometry
def test_perform_calculation_nircam_sw_ts():
    pando = jwst.PandeiaCalculation('nircam', 'sw_ts')
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 10.055)

    # pando.show_config()
    # pando.calc['strategy']
    filter = 'f212n'
    readout = 'rapid'
    subarray = 'sub160p'
    ngroup = 11
    nint = 500

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray, readout=readout,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 10907.382669)
    np.testing.assert_allclose(rep['extracted_flux'], 95399.94145)
    np.testing.assert_allclose(rep['extracted_noise'], 8.746364, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 15469.766)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.640836, rtol=1e-6)


def test_perform_calculation_nircam_lw_ts():
    pando = jwst.PandeiaCalculation('nircam', 'lw_ts')
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 10.055)

    filter = 'f480m'
    readout = 'rapid'
    subarray = 'sub160p'
    ngroup = 11
    nint = 500

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray, readout=readout,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 9885.335196)
    np.testing.assert_allclose(rep['extracted_flux'], 73134.053834)
    np.testing.assert_allclose(rep['extracted_noise'], 7.398237, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 14831.356)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.782422, rtol=1e-6)


def test_perform_calculation_miri_imaging_ts():
    pando = jwst.PandeiaCalculation('miri', 'imaging_ts')
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 10.055)

    filter = 'f560w'
    subarray = 'sub256'
    ngroup = 11
    nint = 500

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 9137.029043)
    np.testing.assert_allclose(rep['extracted_flux'], 156471.800829)
    np.testing.assert_allclose(rep['extracted_noise'], 17.12502, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 39420.387)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.670673, rtol=1e-6)


# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# Acquisition
def test_perform_calculation_miri_target_acq():
    pando = jwst.PandeiaCalculation('miri', 'target_acq')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    filter = 'f1000w'
    subarray = 'slitlessprism'
    ngroup = 10
    nint = 1

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray,
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 512.337046)
    np.testing.assert_allclose(rep['extracted_flux'], 451308.813965)
    np.testing.assert_allclose(rep['extracted_noise'], 880.882648, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 47220.953)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.387804, rtol=1e-6)


def test_perform_calculation_nircam_target_acq():
    pando = jwst.PandeiaCalculation('nircam', 'target_acq')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    filter = 'f335m'
    ngroup = 5
    nint = 1

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, 
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 281.089612)
    np.testing.assert_allclose(rep['extracted_flux'], 2667698.826546)
    np.testing.assert_allclose(rep['extracted_noise'], 9490.563538, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 510198.7)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.483993, rtol=1e-6)


def test_perform_calculation_niriss_target_acq():
    pando = jwst.PandeiaCalculation('niriss', 'target_acq')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    readout = 'nis'
    ngroup = 3
    nint = 1
    report = pando.perform_calculation(ngroup, nint, readout=readout)

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 406.413349)
    np.testing.assert_allclose(rep['extracted_flux'], 469783.817383)
    np.testing.assert_allclose(rep['extracted_noise'], 1155.926148, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 69039.62)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.52355, rtol=1e-6)


def test_perform_calculation_nirspec_target_acq():
    pando = jwst.PandeiaCalculation('nirspec', 'target_acq')
    pando.set_scene('phoenix', 'k2v', 'gaia,g', 14.71)

    filter = 'f110w'
    subarray = 'sub32'
    readout = 'nrsrapid'
    ngroup = 3
    nint = 1

    report = pando.perform_calculation(
        ngroup, nint,
        filter=filter, subarray=subarray, readout=readout
    )

    rep = report['scalar']
    # Differ from ETC because background is not exactly the same
    np.testing.assert_allclose(rep['sn'], 76.970847)
    np.testing.assert_allclose(rep['extracted_flux'], 526950.164062)
    np.testing.assert_allclose(rep['extracted_noise'], 6846.100645, rtol=1e-6)
    np.testing.assert_allclose(rep['brightest_pixel'], 283694.34)
    np.testing.assert_allclose(rep['fraction_saturation'], 0.19588, rtol=1e-6)





import pathlib
import numpy as np
from gen_tso.utils import ROOT
import gen_tso.catalogs.utils as u

# Path to tests folder
ROOT = pathlib.Path(ROOT).parent.joinpath('tests')


def test_get_trexolists_targets_default():
    targets = u.get_trexolists_targets(
        trexo_file=f'{ROOT}/mocks/trexolists.csv',
    )
    expected_targets = [
        'CD-38 2551', 'GJ 341', 'TRAPPIST-1', 'WASP-107', 'WASP-63',
    ]
    np.testing.assert_equal(targets, expected_targets)


def test_get_trexolists_targets_group():
    targets = u.get_trexolists_targets(
        trexo_file=f'{ROOT}/mocks/trexolists.csv',
        grouped=True,
    )
    expected_targets = [
        ['GJ 341'],
        ['WASP-107'],
        ['TRAPPIST-1'],
        ['CD-38 2551', 'WASP-63'],
    ]
    np.testing.assert_equal(targets, expected_targets)


def test_get_trexolists_targets_coords():
    targets, ra, dec = u.get_trexolists_targets(
        trexo_file=f'{ROOT}/mocks/trexolists.csv',
        extract='coords',
    )
    expected_targets = [
        'CD-38 2551', 'GJ 341', 'TRAPPIST-1', 'WASP-107', 'WASP-63',
    ]
    expected_ra = ['06:17:2', '09:21:3', '23:06:3', '12:33:3', '06:17:2']
    expected_dec = ['-38:19', '-60:16', '-05:02', '-10:08', '-38:19']
    np.testing.assert_equal(targets, expected_targets)
    np.testing.assert_equal(ra, expected_ra)
    np.testing.assert_equal(dec, expected_dec)


def test_get_trexolists_targets_group_coords():
    targets, ra, dec = u.get_trexolists_targets(
        trexo_file=f'{ROOT}/mocks/trexolists.csv',
        grouped=True,
        extract='coords',
    )
    expected_targets = [
        ['GJ 341'],
        ['WASP-107'],
        ['TRAPPIST-1'],
        ['CD-38 2551', 'WASP-63'],
    ]
    expected_ra = ['09:21:3', '12:33:3', '23:06:3', '06:17:2']
    expected_dec = ['-60:16', '-10:08', '-05:02', '-38:19']
    np.testing.assert_equal(targets, expected_targets)
    np.testing.assert_equal(ra, expected_ra)
    np.testing.assert_equal(dec, expected_dec)




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import gen_tso.pandeia_io as jwst

import numpy as np
import pytest


instruments = [
    'miri',
    'nircam',
    'niriss',
    'nirspec',
]


def test_get_configs_miri_lrsslitless():
    configs = get_configs(instrument='miri', obs_type='spectroscopy')
    assert len(configs) == 2
    inst = configs[0]
    assert inst['mode'] == 'lrsslitless'
    assert inst['mode_label'] == 'Low Resolution Spectroscopy (LRS) Slitless'


def test_get_configs_miri_mrs_ts():
    configs = get_configs(instrument='miri', obs_type='spectroscopy')
    assert len(configs) == 2
    inst = configs[1]
    assert inst['mode'] == 'mrs_ts'
    assert inst['mode_label'] == 'MRS Time Series'


def test_get_configs_nircam_lw_tsgrism():
    configs = get_configs(instrument='nircam', obs_type='spectroscopy')
    assert len(configs) == 2
    inst = configs[0]
    assert inst['mode'] == 'lw_tsgrism'
    assert inst['mode_label'] == 'LW Grism Time Series'


def test_get_configs_nircam_sw_tsgrism():
    configs = get_configs(instrument='nircam', obs_type='spectroscopy')
    assert len(configs) == 2
    inst = configs[1]
    assert inst['mode'] == 'sw_tsgrism'
    assert inst['mode_label'] == 'SW Grism Time Series'


def test_get_configs_niriss_soss():
    configs = get_configs(instrument='niriss', obs_type='spectroscopy')
    assert len(configs) == 1
    inst = configs[0]
    assert inst['mode'] == 'soss'
    assert inst['mode_label'] == 'SOSS'


def test_get_configs_nirspec_bots():
    configs = get_configs(instrument='nirspec', obs_type='spectroscopy')
    assert len(configs) == 1
    inst = configs[0]
    assert inst['mode'] == 'bots'
    assert inst['mode_label'] == 'Bright Object Time Series'


@pytest.mark.parametrize('instrument', instruments)
def test_get_configs_acquisition(instrument):
    configs = get_configs(instrument=instrument, obs_type='acquisition')
    assert len(configs) == 1
    inst = configs[0]
    assert inst['mode'] == 'target_acq'


def test_load_flux_rate_splines_all():
    # Extract all splines:
    flux_rate_splines, full_wells = jwst.load_flux_rate_splines()
    # Evaluate for one config:
    obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    spline = flux_rate_splines[obs_label]
    flux_rate = spline(8.351)
    full_well = full_wells[obs_label]

    expected_full_well = 58100.001867429375
    expected_flux_rate = 3.1140133020415353
    np.testing.assert_allclose(flux_rate, expected_flux_rate)
    np.testing.assert_allclose(full_well, expected_full_well)


def test_load_flux_rate_splines_single():
    obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    flux_rate_spline, full_well = jwst.load_flux_rate_splines(obs_label)
    flux_rate = flux_rate_spline(8.351)

    expected_full_well = 58100.001867429375
    expected_flux_rate = 3.1140133020415353
    np.testing.assert_allclose(flux_rate, expected_flux_rate)
    np.testing.assert_allclose(full_well, expected_full_well)


def test_load_flux_rate_splines_not_found():
    obs_label = 'nope_lw_tsgrism_f444w_phoenix_k5v'
    flux_rate_spline, full_well = jwst.load_flux_rate_splines(obs_label)
    assert flux_rate_spline is None
    assert full_well is None




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'response_boundaries',
    'plotly_filters',
    'plotly_sed_spectra',
    'plotly_depth_spectra',
    'plotly_tso_spectra',
    'plotly_tso_fluxes',
    'plotly_tso_snr',
    'plotly_tso_2d',
]

from itertools import groupby
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from pyratbay.spectrum import bin_spectrum, constant_resolution_spectrum
from pyratbay.tools import u

from .. import pandeia_io as jwst


COLOR_SEQUENCE = [
    'Royalblue', # blue
    '#15b01a',  # green
    '#000075',  # navy
    '#f032e6',  # magenta
    '#42d4f4',  # cyan
    '#888888',  # grey
    'Red',      # red
    '#9A6324',  # brown
    '#800000',  # maroon
    '#000000',  #  black
    '#469990',  # teal
    '#911eb4',  # purple
    '#808000',  # olive
    'Green',  # green
]

depth_units_label = {
    'none': '',
    'percent': ' (%)',
    'ppm': ' (ppm)',
}


def response_boundaries(wl, response, threshold=0.001):
    """
    Find the wavelength boundaries where a response function
    is greater than the required threshold.

    Parameters
    ----------
    wl: 1D float iterable
        Wavelength array where a response function is sampled
    response: 1D float iterable
        Response function.
    threshold: float
        Minimum response value for flagging.

    Returns
    -------
    bounds: list of float pairs
        A list of the wavelength boundaries for each contiguous
        segment with non-zero response.

    Examples
    --------
    >>> import gen_tso.plotly_io as plots
    >>>
    >>> nwave = 21
    >>> wl = np.linspace(0.0, 1.0, nwave)
    >>> response = np.zeros(nwave)
    >>> response[2:6] = response[10:12] = 1.0
    >>> bounds = plots.response_boundaries(wl, response, threshold=0.5)
    >>> print(bounds)
    [(0.1, 0.25), (0.5, 0.55)]
    """
    bounds = []
    # Contigous ranges where response > threshold:
    for group, indices in groupby(range(len(wl)), lambda x: response[x]>threshold):
        if group:
            indices = list(indices)
            imin = indices[0]
            imax = indices[-1]
            bounds.append((wl[imin], wl[imax]))
    return bounds


def plotly_filters(
        passbands, inst_name, mode_name, subarray_name, filter_name,
        show_all=False,
    ):
    """
    Make a plotly figure of the passband filters.

    Parameters
    ----------
    passbands: Dictionary
        Dictionary of passband arrays. See example below.
    inst_name: String
        Instrument to plot/highlight
    mode_name: String
        Mode to plot/highlight
    subarray: String
        Subarray to plot the passband.
    filter_name: String
        Filter to plot.
    show_all: Bool
        If True, plot a selection from all instruments for better comparison.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> import gen_tso.plotly_io as plots
    >>>
    >>> passbands = jwst.get_throughputs('spectroscopy')
    >>> fig = plots.plotly_filters(
    >>>     passbands, 'nircam', 'lw_tsgrism', 'subgrism64', 'f444w',
    >>>     show_all=True,
    >>> )
    >>> fig.show()
    """
    instruments = [inst_name]
    if show_all:
        instruments += [
            inst for inst in passbands.keys()
            if inst not in instruments
        ]

    show_all_subarrays = {
        'lrsslitless': 'slitlessprism',
        'sw_tsgrism': 'sub40stripe1_dhs',
        'lw_tsgrism': 'subgrism64',
        'bots': 'sub2048',
        'soss': 'substrip256',
    }
    show_all_filters = {
        'lrsslitless': ['None'],
        'lw_tsgrism': ['f322w2', 'f444w'],
        'sw_tsgrism': ['f150w2'],
        'bots': ['g140m/f100lp', 'g235m/f170lp', 'g395m/f290lp', 'prism/clear'],
        'soss': ['clear'],
    }

    # Parse filters to plot
    all_filters = {}
    nfilters = 0
    for inst in instruments:
        modes = list(passbands[inst])
        all_filters[inst] = {}

        for mode in reversed(modes):
            if mode == mode_name and subarray_name in passbands[inst][mode]:
                subarray = subarray_name
                filters = list(passbands[inst][mode][subarray].keys())
            elif show_all and mode in show_all_subarrays:
                subarray = show_all_subarrays[mode]
                filters = show_all_filters[mode]
            else:
                filters = []

            for filter in filters:
                band = passbands[inst][mode][subarray][filter]
                all_filters[inst][filter] = band
                if 'order2' in band:
                    all_filters[inst]['order2'] = band['order2']
        nfilters += len(all_filters[inst])


    visible = [None for _ in range(nfilters)]
    if mode_name == 'bots':
        for i,filter in enumerate(all_filters[inst_name].keys()):
            hide = ('h' in filter_name) is not ('h' in filter)
            if hide and 'prism' not in filter:
                visible[i] = 'legendonly'
    elif mode_name in ['lw_tsgrism', 'sw_tsgrism']:
        nircam_visibles = ['f070w', 'f090w', 'f150w2', 'f322w2', 'f444w']
        for i,filter in enumerate(all_filters[inst_name].keys()):
            if filter != filter_name and filter not in nircam_visibles:
                visible[i] = 'legendonly'

    ncols = len(all_filters[inst_name])
    primary_colors = px.colors.sample_colorscale(
        'Viridis', np.linspace(0, 0.8, ncols),
    )
    secondary_colors = [
        px.colors.sequential.gray[1],
        px.colors.sequential.gray[3],
        px.colors.sequential.gray[5],
        px.colors.sequential.gray[7],
        px.colors.sequential.gray[8],
        px.colors.sequential.gray[9],
    ]
    secondary_nirspec = [
        px.colors.sequential.Magma[3],
        px.colors.sequential.Magma[5],
        px.colors.sequential.Magma[6],
        px.colors.sequential.Magma[7],
    ]
    sel_cols = iter(primary_colors)
    other_cols = iter(secondary_colors)
    nirspec_cols = iter(secondary_nirspec)

    fig = go.Figure()
    j = 0
    filter_index = -1
    for inst in instruments:
        for filter, throughput in all_filters[inst].items():
            if filter == filter_name:
                color = next(sel_cols)
                color = 'Gold'
                width = 3.5
                filter_index = j
            elif filter == 'order2' and filter_name == 'clear':
                color = 'Orange'
                width = 3.5
            elif inst == inst_name:
                color = next(sel_cols)
                width = 2.0
            else:
                color = next(nirspec_cols) if inst=='nirspec' else next(other_cols)
                width = 1.25
            linedict = dict(color=color, width=width)
            fig.add_trace(go.Scatter(
                x=throughput['wl'],
                y=throughput['response'],
                mode='lines',
                name=filter.upper(),
                legendgroup=inst,
                legendgrouptitle_text=inst,
                line=linedict,
                legendrank=j+nfilters*int(inst != inst_name),
                visible=visible[j],
            ))
            j += 1
    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'throughput = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text=None,
        autorangeoptions=dict(minallowed=0),
    )

    wl_scale = 'log'
    wl_range = [np.log10(0.6), np.log10(13.5)]
    wl_range = [np.log10(0.6), np.log10(28.0)]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )
    fig.update_layout(
        showlegend=True,
        legend=dict(groupclick="toggleitem"),
    )

    # Show current filter on top (TBD: there must be a better way)
    if filter_index < 0:
        return fig
    fig_idx = np.arange(len(fig.data))
    fig_idx[-1] = filter_index
    fig_idx[filter_index] = len(fig.data) - 1
    fig.data = tuple(np.array(fig.data)[fig_idx])
    return fig


def plotly_sed_spectra(
        sed_models, labels, highlight_model=None,
        wl_range=None, units='mJy', wl_scale='linear', resolution=250.0,
        throughput=None,
    ):
    """
    Make a plotly figure of stellar SED spectra.
    """
    nmodels = len(sed_models)
    fig = go.Figure(
        layout={'colorway': COLOR_SEQUENCE},
    )

    # Shaded area for filter:
    if throughput is not None:
        if 'order2' in throughput:
            wl = throughput['order2']['wl']
            response = throughput['order2']['response']
            band_bounds = response_boundaries(wl, response, threshold=0.03)
            for bound in band_bounds:
                fig.add_vrect(
                    fillcolor="orchid", opacity=0.4,
                    x0=bound[0], x1=bound[1],
                    layer="below", line_width=0,
                )
        band_bounds = response_boundaries(throughput['wl'], throughput['response'])
        for bound in band_bounds:
            fig.add_vrect(
                fillcolor="#069af3", opacity=0.4,
                x0=bound[0], x1=bound[1],
                layer="below", line_width=0,
            )

    for j,model in enumerate(sed_models):
        wl = model['wl']
        flux = model['flux']
        if resolution > 0:
            wl_min = np.amin(wl)
            wl_max = np.amax(wl)
            bin_wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            bin_flux = bin_spectrum(bin_wl, wl, flux, gaps='interpolate')
            mask = np.isfinite(bin_flux)
            wl = bin_wl[mask]
            flux = bin_flux[mask]

        if highlight_model is None:
            linedict = dict(width=1.25)
            rank = j
        elif labels[j] == highlight_model:
            linedict = dict(color='Gold', width=2.0)
            rank = j + nmodels
        else:
            linedict = dict(width=1.25)
            rank = j
        fig.add_trace(go.Scatter(
            x=wl,
            y=flux,
            mode='lines',
            opacity=0.75,
            name=labels[j],
            line=linedict,
            legendrank=rank,
        ))

    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'flux = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text=f'Flux ({units})',
        title_standoff=0,
    )

    if wl_scale == 'log':
        wl_range = [
            None if wave is None else np.log10(wave)
            for wave in wl_range
        ]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


def plotly_depth_spectra(
        depth_models, labels, highlight_model=None,
        wl_range=None, units='percent', wl_scale='linear', resolution=250.0,
        depth_range=None,
        obs_geometry='transit',
        throughput=None,
    ):
    """
    Make a plotly figure of transit/eclipse depth spectra.
    """
    if depth_range is None:
        depth_range =[None, None]

    nmodels = len(depth_models)
    fig = go.Figure(
        layout={'colorway':COLOR_SEQUENCE},
    )
    # Shaded area for filter:
    if throughput is not None:
        if 'order2' in throughput:
            wl = throughput['order2']['wl']
            response = throughput['order2']['response']
            band_bounds = response_boundaries(wl, response, threshold=0.03)
            for bound in band_bounds:
                fig.add_vrect(
                    fillcolor="orchid", opacity=0.4,
                    x0=bound[0], x1=bound[1],
                    layer="below", line_width=0,
                )
        bounds = response_boundaries(throughput['wl'], throughput['response'])
        for bound in bounds:
            fig.add_vrect(
                fillcolor="#069af3", opacity=0.4,
                x0=bound[0], x1=bound[1],
                layer="below", line_width=0,
            )

    ymax = 0.0
    ymin = np.inf
    for j,model in enumerate(depth_models):
        wl = model['wl']
        depth = model['depth']
        if resolution > 0:
            wl_min = np.amin(wl)
            wl_max = np.amax(wl)
            bin_wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            depth = bin_spectrum(bin_wl, wl, depth, gaps='interpolate')/u(units)
            mask = np.isfinite(depth)
            wl = bin_wl[mask]
            depth = depth[mask]

        if labels[j] == highlight_model:
            linedict = dict(color='Gold', width=2.0)
            rank = j + nmodels
        else:
            linedict = dict(width=1.25)
            rank = j
        fig.add_trace(go.Scatter(
            x=wl,
            y=depth,
            mode='lines',
            name=labels[j],
            line=linedict,
            legendrank=rank,
        ))
        ymax = np.amax([ymax, np.amax(depth)])
        ymin = np.amin([ymin, np.amin(depth)])


    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'depth = %{y:.3f}'
    )
    if depth_range[0] is None or depth_range[1] is None:
        dy = 0.05 * (ymax-ymin)
    if depth_range[0] is None:
        depth_range[0] = ymin - dy
    if depth_range[1] is None:
        depth_range[1] = ymax + dy

    ylabel = f'{obs_geometry} depth{depth_units_label[units]}'
    fig.update_yaxes(
        title_text=ylabel,
        title_standoff=0,
        range=depth_range,
    )

    if wl_scale == 'log' and wl_range is not None:
        wl_range = [
            None if wave is None else np.log10(wave)
            for wave in wl_range
        ]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


def plotly_tso_spectra(
        tso_list, sim_depths=None, resolution=250.0, n_obs=1,
        model_label='model', instrument_label=None,
        units='percent', wl_range=None, wl_scale='linear',
        depth_range=None,
        obs_geometry='transit',
    ):
    """
    Make a plotly figure of transit/eclipse depth TSO spectra.
    """
    if not isinstance(tso_list, list):
        tso_list = [tso_list]

    if isinstance(instrument_label, str):
        instrument_label = [instrument_label for tso in tso_list]

    fig = go.Figure()
    obs_col = px.colors.sample_colorscale('Viridis', 0.2)[0]
    model_col = px.colors.sample_colorscale('Viridis', 0.75)[0]

    ymax = 0.0
    ymin = np.inf
    legends = []
    for i,tso in enumerate(tso_list):
        if sim_depths is None:
            bin_wl, bin_spec, bin_err, wl_err = jwst.simulate_tso(
               tso, n_obs=n_obs, resolution=resolution, noiseless=False,
            )
        else:
            bin_wl = sim_depths[i]['wl']
            bin_spec = sim_depths[i]['depth']
            bin_err = sim_depths[i]['uncert']
            wl_err = sim_depths[i]['wl_widths']

        mode = tso['report_in']['input']['configuration']['instrument']['mode']
        if mode in jwst._photo_modes:
            input_wl, input_depth = tso['input_depth']
            wl_min = np.amin(input_wl)
            wl_max = np.amax(input_wl)
            wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            spec = bin_spectrum(wl, input_wl, input_depth, gaps='interpolate')
        else:
            wl = tso['wl']
            spec = tso['depth_spectrum']

        show_legend = model_label not in legends
        fig.add_trace(go.Scatter(
            x=wl,
            y=spec/u(units),
            mode='lines',
            name=model_label,
            legendgroup=model_label,
            showlegend=show_legend,
            line=dict(color=model_col, width=1.5),
        ))
        legends.append(model_label)

        if mode in jwst._photo_modes:
            error_x = dict(
                type='data', symmetric=False, visible=True,
                array=wl_err[:,1], arrayminus=wl_err[:,0],
            )
        else:
            error_x = None

        show_legend = instrument_label[i] not in legends
        fig.add_trace(go.Scatter(
            x=bin_wl,
            y=bin_spec/u(units),
            error_y=dict(type='data', array=bin_err/u(units), visible=True),
            error_x=error_x,
            mode='markers',
            name=instrument_label[i],
            legendgroup=instrument_label[i],
            showlegend=show_legend,
            marker=dict(color=obs_col, size=5),
        ))
        legends.append(instrument_label[i])
        ymax = np.amax([ymax, np.amax(spec)])
        ymin = np.amin([ymin, np.amin(spec)])

    # Saturation (take report with highest e-/sec)
    report = tso['report_out'] if obs_geometry=='transit' else tso['report_in']
    wl, partial = report['1d']['n_partial_saturated']
    wl, full = report['1d']['n_full_saturated']
    partial_saturation = response_boundaries(wl, partial, threshold=0)
    for j,bound in enumerate(partial_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="red", opacity=0.3,
            layer="below", line_width=0,
            legendgrouptitle_text="Saturation",
            legendgroup='saturation',
            name='partial',
            showlegend=(j==0),
        )
    full_saturation = response_boundaries(wl, full, threshold=0)
    for j,bound in enumerate(full_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="black", opacity=0.75,
            layer="below", line_width=0,
            legendgroup='saturation',
            name='full',
            showlegend=(j==0),
        )

    fig.update_traces(
        hovertemplate=
            'wl = %{x:.2f}<br>'+
            'depth = %{y:.3f}'
    )
    if depth_range is None:
        ymax = ymax/u(units)
        ymin = ymin/u(units)
        dy = 0.1 * (ymax-ymin)
        depth_range = [ymin-dy, ymax+dy]
    ylabel = f'{obs_geometry} depth{depth_units_label[units]}'
    fig.update_yaxes(
        title_text=ylabel,
        title_standoff=0,
        range=depth_range,
    )

    if wl_scale == 'log' and wl_range is not None:
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        entrywidth=1.0,
        entrywidthmode='fraction',
        yanchor="bottom",
        xanchor="right",
        y=1.02,
        x=1
    ))
    fig.update_layout(showlegend=True)
    return fig


def plotly_tso_fluxes(
        tso_list,
        wl_range=None, wl_scale='linear',
        obs_geometry='transit',
    ):
    """
    Plot 1D source and background flux rates
    """
    if not isinstance(tso_list, list):
        tso_list = [tso_list]

    colors = [
        px.colors.sample_colorscale('Viridis', 0.15)[0],
        px.colors.sample_colorscale('Viridis', 0.7)[0],
        px.colors.sample_colorscale('Viridis', 0.9)[0],
    ]
    legends = ['in-transit', 'out-transit', 'background']

    fig = go.Figure()
    for j,tso in enumerate(tso_list):
        wl = tso['report_in']['1d']['extracted_flux'][0]
        fluxes = [
            tso['report_in']['1d']['extracted_flux'][1],
            tso['report_out']['1d']['extracted_flux'][1],
            tso['report_out']['1d']['extracted_bg_only'][1],
        ]
        show_legend = j == 0
        for i in range(len(fluxes)):
            fig.add_trace(go.Scatter(
                x=wl,
                y=fluxes[i],
                mode='lines',
                line=dict(color=colors[i], width=1.75),
                name=legends[i],
                legendgroup=legends[i],
                showlegend=show_legend,
            ))

    # Saturation (take report with highest e-/sec)
    report = tso['report_out'] if obs_geometry=='transit' else tso['report_in']
    wl, partial = report['1d']['n_partial_saturated']
    wl, full = report['1d']['n_full_saturated']
    partial_saturation = response_boundaries(wl, partial, threshold=0)
    for j,bound in enumerate(partial_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="red", opacity=0.3,
            layer="below", line_width=0,
            legendgrouptitle_text="Saturation",
            name='partial',
            legendgroup='saturation',
            showlegend=(j==0),
        )
    full_saturation = response_boundaries(wl, full, threshold=0)
    for j,bound in enumerate(full_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="black", opacity=0.75,
            layer="below", line_width=0,
            name='full',
            legendgroup='saturation',
            showlegend=(j==0),
        )

    fig.update_traces(
        hovertemplate='wl = %{x:.2f}<br>' + 'flux = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text='flux rate (e-/sec)',
        title_standoff=0,
    )
    if wl_scale == 'log' and wl_range is not None:
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )
    fig.update_layout(showlegend=True)
    return fig


def plotly_tso_snr(
        tso_list,
        wl_range=None, wl_scale='linear',
        obs_geometry='transit',
    ):
    """
    Plot 1D signal-to-noise ratios for in- and out-of-transit simulations
    """
    if not isinstance(tso_list, list):
        tso_list = [tso_list]

    colors = [
        px.colors.sample_colorscale('Viridis', 0.15)[0],
        px.colors.sample_colorscale('Viridis', 0.7)[0],
    ]
    legends = ['in-transit', 'out-transit']

    fig = go.Figure()
    for j,tso in enumerate(tso_list):
        wl = tso['report_in']['1d']['sn'][0]
        snr = [
            tso['report_in']['1d']['sn'][1],
            tso['report_out']['1d']['sn'][1],
        ]
        show_legend = j == 0
        for i in range(len(snr)):
            fig.add_trace(go.Scatter(
                x=wl,
                y=snr[i],
                mode='lines',
                line=dict(color=colors[i], width=1.75),
                name=legends[i],
                legendgroup=legends[i],
                showlegend=show_legend,
            ))

    # Saturation (take report with highest e-/sec)
    report = tso['report_out'] if obs_geometry=='transit' else tso['report_in']
    wl, partial = report['1d']['n_partial_saturated']
    wl, full = report['1d']['n_full_saturated']
    partial_saturation = response_boundaries(wl, partial, threshold=0)
    for j,bound in enumerate(partial_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="red", opacity=0.3,
            layer="below", line_width=0,
            legendgrouptitle_text="Saturation",
            name='partial',
            legendgroup='saturation',
            showlegend=(j==0),
        )
    full_saturation = response_boundaries(wl, full, threshold=0)
    for j,bound in enumerate(full_saturation):
        fig.add_vrect(
            x0=bound[0], x1=bound[1],
            fillcolor="black", opacity=0.75,
            layer="below", line_width=0,
            name='full',
            legendgroup='saturation',
            showlegend=(j==0),
        )

    fig.update_traces(
        hovertemplate='wl = %{x:.2f}<br>' + 'flux = %{y:.3f}'
    )
    fig.update_yaxes(
        title_text='signal-to-noise ratio',
        title_standoff=0,
    )
    if wl_scale == 'log' and wl_range is not None:
        wl_range = [np.log10(wave) for wave in wl_range]
    fig.update_xaxes(
        title_text='wavelength (um)',
        title_standoff=0,
        range=wl_range,
        type=wl_scale,
    )
    fig.update_layout(showlegend=True)
    return fig


def plotly_tso_2d(tso, heatmap_name):
    """
    Make 2D pandeia plots as in the ETC.

    Parameters
    ----------
    tso: Dictionary
        A TSO output.
    heatmap_name: String
        Heat map property to plot, select from:
        'snr', 'detector', 'saturation', or 'ngroups_map'
    """
    # TBD: Think how to multipanel MRS heatmaps
    if isinstance(tso, list):
        report = tso[0]['report_out']
    else:
        report = tso['report_out']
    inst = report['input']['configuration']['instrument']['instrument']
    mode = report['input']['configuration']['instrument']['mode']
    heatmap = report['2d'][heatmap_name]
    # Shiny can't handle nans
    is_nan = ~np.isfinite(heatmap)
    heatmap[is_nan] = 0.0


    x_min = report['transform']['x_min']
    x_max = report['transform']['x_max']
    nx = report['transform']['x_size']
    x = np.linspace(x_min, x_max, nx)
    xstep = report['transform']['x_step']

    y_min = report['transform']['y_min']
    y_max = report['transform']['y_max']
    ny = report['transform']['y_size']
    y = np.linspace(y_min, y_max, ny)
    ystep = report['transform']['y_step']

    if inst == 'niriss':
        nx = report['transform']['x_size']
        ny = report['transform']['y_size']
        xlabel = ylabel = 'pixels'
        heatmap = np.flipud(heatmap)
    elif mode == 'mrs_ts':
        xlabel = 'arcsec'
        ylabel = 'arcsec'
    elif mode == 'lrsslitless':
        y = np.flip(y)
        xlabel = 'dispersion (arcsec)'
        ylabel = 'wavelength (arcsec)'
    else:
        x = report['1d']['sn'][0]
        xlabel = 'wavelength (um)'
        ylabel = 'dispersion (arcsec)'

    # Strategy:
    apertures = []
    if inst != 'niriss':
        aperture = report['input']['strategy']['aperture_size']
        annulus = report['input']['strategy']['sky_annulus']
        apertures = [
            np.tile( 0.5*aperture, 2),
            np.tile(-0.5*aperture, 2),
            np.tile( annulus[0], 2),
            np.tile(-annulus[0], 2),
            np.tile( annulus[1], 2),
            np.tile(-annulus[1], 2),
        ]

    fig = go.Figure(
        data=go.Heatmap(z=heatmap, x=x, y=y, showscale=False),
    )
    if mode == 'mrs_ts':
        t = np.linspace(0.0, 2.0*np.pi, 100)
        for i, aper in enumerate(apertures):
            if i%2 == 1:
                continue
            xx = aper[0]*np.sin(t)
            yy = aper[0]*np.cos(t)
            if i < 2:
                name = 'aperture'
                color= 'gold'
                dash = None
            else:
                name = 'background'
                color= 'limegreen'
                dash = 'dash'
            showlegend = i in [0,2]
            fig.add_trace(go.Scatter(
                x=xx,
                y=yy,
                mode='lines',
                line=dict(color=color, width=2.0, dash=dash),
                name=name, legendgroup=name,
                showlegend=showlegend,
            ))
        fig.update_layout(
            yaxis=dict(scaleanchor='x', scaleratio=1),
        )

    elif inst != 'niriss':
        for i, aper in enumerate(apertures):
            if i < 2:
                name = 'aperture'
                color= 'gold'
                dash = None
            else:
                name = 'background'
                color= 'limegreen'
                dash = 'dash'
            showlegend = i in [0,2]
            if mode == 'lrsslitless':
                xx = aper
                yy = [np.amin(y), np.amax(y)]
            else:
                xx = [np.amin(x), np.amax(x)]
                yy = aper
            fig.add_trace(go.Scatter(
                x=xx,
                y=yy,
                mode='lines',
                line=dict(color=color, width=2.0, dash=dash),
                name=name, legendgroup=name,
                showlegend=showlegend,
            ))

    fig.update_yaxes(
        title_text=ylabel,
        title_standoff=0,
    )
    fig.update_xaxes(
        title_text=xlabel,
        title_standoff=0,
    )

    fig.update_layout(legend=dict(
        orientation="h",
        yanchor="bottom",
        y=1.02,
        xanchor="center",
        x=0.5
    ))

    return fig





# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    # To fetch the JWST programs info from the internets
    '_extract_html_text',
    '_fetch_programs',
    '_extract_instrument_template',
    'fetch_jwst_programs',
    'update_jwst_programs',
    # To format XML info into human language
    '_parse_date',
    '_parse_window',
    '_get_phase_info',
    'guess_event_type',
    'parse_status',
    'parse_program',
    # To guess the planet in an observation
    '_planet_from_label',
    '_planet_from_period',
    '_clean_label',
    'get_observation_host',
    'get_planet_letters',
]

from concurrent.futures import ThreadPoolExecutor
import csv
from datetime import datetime
import os
import re
import shutil
import warnings
import xml.etree.ElementTree as ET

import numpy as np
from bs4 import BeautifulSoup
import pyratbay.constants as pc
import requests

from ..utils import ROOT, KNOWN_PROGRAMS
from .utils import (
    normalize_name,
    get_host,
    get_letter,
)
from .catalogs import load_programs, Catalog


def _extract_html_text(html):
    """
    Parse an HTML file, extract content inside its <main> tag
    """
    soup = BeautifulSoup(html, "html.parser")
    main_tag = soup.find("main")
    return main_tag.get_text(separator=" ", strip=True) if main_tag else None


def _fetch_programs(programs, output_path, file_type='APT'):
    """
    Fetch APT files or status XML files from the Space Telescope website.

    Parameters
    ----------
    programs: 1D integer iterable.
        Program IDs.
    output_path: String
        Output folder location.
    file_type: String
        'APT' or 'status'.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> programs = [1366, 1729, 1981, 2055, 9999]
    >>> output_path = f'{ROOT}data/programs/'
    >>> file_type = 'APT'
    >>> cat._fetch_programs(programs, output_path, file_type)
    """
    nprograms = len(programs)

    if file_type == 'APT':
        urls = np.array([
            f'https://www.stsci.edu/jwst-program-info/download/jwst/apt/{pid}/'
            for pid in programs
        ])
    elif file_type == 'status':
        urls = np.array([
            f'https://www.stsci.edu/jwst-program-info/visits/?program={pid}&download=&pi=1'
            for pid in programs
        ])

    def fetch_url(url):
        try:
            response = requests.get(url)
            return response
        except:
            return None

    fetch_status = np.tile(2, nprograms)
    responses = np.tile({}, nprograms)
    n_attempts = 0
    while np.any(fetch_status>0) and n_attempts < 10:
        n_attempts += 1
        mask = fetch_status > 0
        with ThreadPoolExecutor() as executor:
            results = list(executor.map(fetch_url, urls[mask]))

        j = 0
        for i in range(nprograms):
            if fetch_status[i] <= 0:
                continue
            r = results[j]
            j += 1
            if r is None:
                continue
            if not r.ok:
                warnings.warn(
                    f"{file_type} fetch failed for program {programs[i]} "
                    f"with error {r.status_code}: {repr(r.reason)}"
                )
                fetch_status[i] = -1
                continue
            if file_type=='status' and 'xml' not in r.headers['Content-Type']:
                warnings.warn(_extract_html_text(r.text))
                fetch_status[i] = -1
                continue
            responses[i] = r.content
            fetch_status[i] = 0
        fetched = np.sum(fetch_status <= 0)
        print(f'Fetched {fetched}/{nprograms} {file_type} files on try {n_attempts}')

    for i,resp in enumerate(responses):
        if fetch_status[i] != 0:
            continue
        pid = programs[i]
        if file_type == 'APT':
            file_path = f'{output_path}/{pid}.aptx'
        elif file_type == 'status':
            file_path = f'{output_path}/JWST-{pid}-visit-status.xml'
        with open(file_path, mode="wb") as file:
            file.write(resp)
    return np.array(programs)[fetch_status==0]


def fetch_jwst_programs(programs, apt_command=None, output_path=None):
    r"""
    Fetch APT and status for requested programs.
    APT files will be converted to XML using apt_command.

    Parameters
    ----------
    programs: integer or 1D iterable of intergers
        The program PID numbers to fetch.
    apt_command: String
        APT's command executable.
        If None, do not convert .aptx files to .xml
    output_path: String
        Output folder location.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> from gen_tso.utils import KNOWN_PROGRAMS
    >>>
    >>> # Fetch all known TSO programs
    >>> apt_command = '/Applications/APT\\ 2025.1/bin/apt'
    >>> programs = KNOWN_PROGRAMS
    >>> cat.fetch_jwst_programs(programs, apt_command)
    >>>
    >>> # Fetch a specific TSO program
    >>> programs = 1366
    >>> cat.fetch_jwst_programs(programs, apt_command)
    """
    if isinstance(programs, str) or isinstance(programs, int):
        programs = [programs]

    if output_path is None:
        output_path = f'{ROOT}data/programs/'

    # Fetch status xml files
    _fetch_programs(programs, output_path, 'status')

    # Fetch APT files, export to xml
    apt_programs = _fetch_programs(programs, output_path, 'APT')
    if apt_command is not None:
        if apt_command in os.environ:
           apt_command = os.environ[apt_command]
        apt_files = " ".join(
            [f'{output_path}/{pid}.aptx' for pid in apt_programs]
        )
        os.system(f"{apt_command} -nogui -export xml {apt_files}")


def update_jwst_programs():
    """
    Update JWST programs, re-processing only the necessary APTs
    """
    output_path = f'{ROOT}data/programs/'
    programs = KNOWN_PROGRAMS
    to_csv = f'{ROOT}data/programs/jwst_tso_programs.csv'
    failed = ['Failed', 'Skipped', 'Withdrawn']

    if 'APT' not in os.environ:
        # Only update status
        _fetch_programs(programs, output_path, 'status')
    else:
        # Fetch status in tmp folder
        apt_command = os.environ['APT']
        tmp_path = f'{ROOT}data/programs_tmp/'
        _fetch_programs(programs, tmp_path, 'status')

        # Find observations that failed since last update
        programs = load_programs()
        pid = ''
        to_update = []
        for obs in programs:
            if obs['pid'] != pid:
                pid = obs['pid']
                status = parse_status(pid, path=tmp_path)
            for i in range(len(status)):
                if status[i]['observation'] == obs['observation']:
                    break
            state = status[i]
            if state['status'] != obs['status'] and state['status'] in failed:
                to_update.append(pid)

        # Fetch APTs of programs with failed observations
        if len(to_update) > 0:
            fetch_jwst_programs(to_update, apt_command, output_path)
        # Move status xml files
        for file in os.listdir(tmp_path):
            if not file.endswith('-visit-status.xml'):
                continue
            src = os.path.join(tmp_path, file)
            dest = os.path.join(output_path, file)
            shutil.move(src, dest)

    # Parse all
    parse_program(programs, to_csv=to_csv)


def _extract_instrument_template(element):
    """
    Extract the instrumental settings from an observation's XML template.
    """
    modes = {
        'MiriImaging': 'Imaging TS',
        'MiriLRS': 'LRS',
        'MiriMRS': 'MRS',
        'NircamGrismTimeSeries': 'GRISMR TS',
        'NirissSoss': 'SOSS',
        'NirspecBrightObjectTimeSeries': 'BOTS',
    }

    tag = element.tag
    len_ns = tag.rindex('}') + 1
    xmlns = tag[0:len_ns]
    mode = tag[len_ns:]
    template = {
        'mode': modes[mode],
    }
    if template['mode'] == 'BOTS':
        template['disperser'] = element.find(f".//{xmlns}Grating").text
    elif template['mode'] == 'MRS':
        template['disperser'] = element.find(f".//{xmlns}Wavelength").text
    else:
        template['disperser'] = 'None'

    if template['mode'] == 'GRISMR TS':
        sw_filter = element.find(f".//{xmlns}ShortPupilFilter").text
        sw_disperser, sw_filter = sw_filter.split('+')
        sw_disperser = sw_disperser.replace('GDHS0', 'DHS0')
        lw_filter = element.find(f".//{xmlns}LongPupilFilter").text
        lw_disperser, lw_filter = lw_filter.split('+')
        template['disperser'] = f'{sw_disperser},{lw_disperser}'
        template['filter'] = f'{sw_filter},{lw_filter}'
    elif template['mode'] in ['Imaging TS', 'BOTS']:
        template['filter'] = element.find(f".//{xmlns}Filter").text
    elif template['mode'] in ['SOSS']:
        template['filter'] = 'CLEAR'
    else:
        template['filter'] = 'None'

    template['subarray'] = element.find(f".//{xmlns}Subarray").text
    if template['mode'] == 'MRS':
        readout = element.find(f".//{xmlns}ReadoutPatternLong").text
        groups = element.find(f".//{xmlns}GroupsLong").text
        integs = element.find(f".//{xmlns}IntegrationsLong").text
    else:
        readout = element.find(f".//{xmlns}ReadoutPattern").text
        groups = element.find(f".//{xmlns}Groups").text
        integs = element.find(f".//{xmlns}Integrations").text
    template['readout'] = readout
    template['groups'] = int(groups)
    template['integrations'] = int(integs)

    return template


def _parse_date(date_text):
    """
    Parse date format from a status' start/end date.

    Returns
    -------
    date:
        A datetime.datetime object
    """
    date_format = "%b %d, %Y %H:%M:%S"
    try:
        date = datetime.strptime(date_text, date_format)
    except:
        date = date_text
    return date


def _parse_window(window_text):
    """
    Parse date format from a status' start/end date.

    Returns
    -------
    window: String
        A string representation of the start - end date span
    """
    if "(" not in window_text:
        return "not yet assigned"

    idx_end = window_text.index('(')
    dates = window_text[0:idx_end].split('-')

    window_format = "%b %d, %Y"
    start = datetime.strptime(dates[0].strip(), window_format)
    end = datetime.strptime(dates[1].strip(), window_format)
    start_date = start.strftime('%Y-%m-%d')
    end_date = end.strftime('%Y-%m-%d')
    window = f"{start_date} - {end_date}"
    return window


def _get_phase_info(obs):
    """
    Extract the orbital period, starting phase, and duration
    from an observations from the PeriodZeroPhase constraints.

    Note that output values are adjusted to reflect the astrophysical
    values, since some constraints may have been modified to ease the
    scheduling.

    Parameters
    ----------
    obs: Dictionary
        A JWST observation dict.

    Returns
    -------
    period: Float
        Orbital period in days.
    phase: Float
        Orbital phase at start of observation (midpoint between
        PhaseStart and PhaseEnd).
    phase_duration: Float
        Duration of the observation in orbital phase units.
    """
    if obs['phase_reqs'] is None:
        return None, None, None
    phase_reqs = obs['phase_reqs']

    # The orbital period in days:
    value, unit = phase_reqs['Period'].split()
    per_unit = unit.lower().rstrip('s')
    period = float(value) * getattr(pc, per_unit) / pc.day

    duration = obs['duration']
    phase_duration = (duration*pc.hour) / (period*pc.day)

    # The mid-point of the starting phase range:
    phase_start = float(phase_reqs['PhaseStart'])
    phase_end = float(phase_reqs['PhaseEnd'])
    phase = 0.5*(phase_start+phase_end)
    if phase < 0:
         phase += 1

    # Phase curves have durations longer than 1 orbit,
    # but phase-curved have halved-period values to ease scheduling
    is_phase = phase_duration > 0.54
    phase_doubling = 1
    pid = obs['pid']
    if pid == '3860':
        phase_doubling = 4
    elif is_phase or pid in ['2084']:
        phase_doubling = 2

    phase_duration *= phase_doubling
    period /= phase_doubling
    phase *= phase_doubling
    # Manual guesstimate
    if pid == '1201' and 'eclipse' in obs['label']:
        phase -= 0.5
    phase -= int(phase)

    return period, phase, phase_duration


def guess_event_type(obs):
    """
    Use labels or phase constraints to guess whether the observation
    is an eclipse, transit, or phase curve.

    If there's orbital phase information, it will
    """
    pid = obs['pid']
    label = obs['label']
    event = ''

    # Guess from orbital phase when phase constraints exist:
    if obs['period'] is not None:
        phase = obs['phase_start']
        duration = obs['phase_duration']
        if duration > 1.0:
            event = 'phase curve'
        elif (phase<1.0) and (phase+duration>1.0):
            event = 'transit'
        else:
            event = 'eclipse'

    # Use label (can override phase-guess, and that's intentional):
    if 'trans' in label:
        event = 'transit'
    elif 'phase' in label:
        event = 'phase curve'
    elif 'eclipse' in label or 'emis' in label or 'occultat' in label:
        event = 'eclipse'

    # Hardcoded patches for missing information:
    if pid in ['2149', '2589', '3385', '5177', '5882', '6456']:
        event = 'transit'
    elif pid in ['2488', '2765']:
        event = 'phase curve'

    if event == '':
        obs_id = obs['observation']
        visit = obs['visit']
        print(f'{pid} {obs_id:3} {visit}  Could not identify event type')
    return event


def parse_status(pid, path=None):
    """
    Parse a program's status xml file to list of dictionaries

    Parameters
    ----------
    pid: Integer
        The programs PID number.
    path: String
        Folder where the status xml file is loacted.

    Returns
    -------
    visit_status: List of dictionaries
        Dictionaries with the observations' status for given program
    """
    if path is None:
        path = f'{ROOT}data/programs/'

    status = ET.parse(f'{path}/JWST-{pid}-visit-status.xml').getroot()
    visit_status = []
    for visit in status.findall('.//visit'):
        state = visit.attrib
        state['status'] = visit.find('status').text
        state['duration'] = float(visit.find('hours').text)
        target = visit.find('target')
        if target is None:
            target = 'None'
        else:
            target = target.text
        start = visit.find('startTime')
        end = visit.find('endTime')
        window = visit.find('planWindow')

        state['date_start'] = None
        state['date_end'] = None
        state['plan_window'] = None
        if start is not None:
            state['date_start'] = _parse_date(start.text)
        if end is not None:
            state['date_end'] = _parse_date(end.text)
        if window is not None:
            state['plan_window'] = _parse_window(window.text)
        visit_status.append(state)
    return visit_status


def parse_program(pid, path=None, to_csv=None):
    """
    Parse a JWST program's xml files (of APT and status) to extract
    the program's, status, and observation information.

    Parameters
    ----------
    pid: Integer or 1D list of integers.
        JWST TSO program ID number(s).
    path: String
        Path where the input XML files are located. These are
        files downloaded with cat.fetch_jwst_programs().
    to_csv: string
        If not None, save the outputs to a csv file located in
        given filename path.

    Returns
    -------
    observations: List of dictionaries
        Dictionaries with the observations information:
        - The program's category, PI, PID, cycle, and proprietary period
        - The target's name, ra, and dec
        - The observation's number, visit, duration (hours),
          status, start and end dates (or planned window), labels,
          and event type (eclipse, transit, or phase curve)
        - The instrument, detector mode, disperser, filter,
          subarray, readout, groups, and integrations.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> obs = cat.parse_program(pid=3712)

    >>> import gen_tso.catalogs as cat
    >>> obs = cat.parse_program(pid=3712, to_csv='jwst_tso_program_3712.csv')

    >>> import gen_tso.catalogs as cat
    >>> from gen_tso.utils import ROOT, KNOWN_PROGRAMS
    >>> # Save to default csv file
    >>> to_csv = f'{ROOT}data/programs/jwst_tso_programs.csv'
    >>> observations = cat.parse_program(pid=KNOWN_PROGRAMS, to_csv=to_csv)
    """
    if path is None:
        path = f'{ROOT}data/programs/'

    if isinstance(pid, (int, str)):
        pid = [pid]

    observations = []
    for program in pid:
        status = parse_status(program, path)
        # APT
        root = ET.parse(f'{path}/{program}.xml').getroot()
        ns = {'apt': 'http://www.stsci.edu/JWST/APT'}

        # The program
        pi_category = root.find('.//apt:ProposalCategory', ns).text
        cycle = root.find('.//apt:Cycle', ns).text
        prop = root.find('.//apt:ProprietaryPeriod', ns).text
        proprietary_period = prop[prop.find('[')+1:prop.find(' Month')]
        pi = ".//apt:PrincipalInvestigator/apt:InvestigatorAddress/apt:LastName"
        pi_lastname = root.find(pi, ns).text

        # The targets
        targets = {}
        for child in root.findall('.//apt:Targets/apt:Target', ns):
            target = {}
            target['number'] = child.find('apt:Number', ns).text
            target['ID'] = child.find('apt:TargetID', ns).text
            target['name'] = child.find('apt:TargetName', ns).text
            archive_name = child.find('apt:TargetArchiveName', ns)
            if archive_name is not None:
                target['archive_name'] = archive_name.text
            coords = child.find('apt:EquatorialCoordinates', ns)
            if coords is None:
                continue
            coords = coords.get('Value')
            target['ra'] = ':'.join(coords.split()[0:3])
            target['dec'] = ':'.join(coords.split()[3:6])
            epoch = child.find('apt:Epoch', ns)
            if epoch is not None:
                target['epoch'] = epoch.text
            targets[f"{target['number']} {target['ID']}"] = target

        # The observations
        for obs_group in root.findall( './/apt:ObservationGroup', ns):
            label = obs_group.find('apt:Label', ns)
            group_label = "" if label is None else label.text
            for obs in obs_group.findall('.//apt:Observation', ns):
                observation = {
                    'category': pi_category,
                    'pid': str(program),
                    'pi': pi_lastname,
                    'cycle': int(cycle),
                }
                reqs = np.unique([
                    child.tag[child.tag.rindex('}')+1:]
                    for child in obs.find(".//apt:SpecialRequirements", ns)
                ]).tolist()
                phase_reqs = obs.find(".//apt:PeriodZeroPhase", ns)
                time_series_reqs = obs.find(".//apt:TimeSeriesObservation", ns)
                if time_series_reqs is None:
                    continue

                observ = obs.find('apt:Number', ns).text
                visit = obs.find('apt:Visit', ns).get('Number')
                for state in status:
                    if state['observation'] == observ and state['visit'] == visit:
                        break
                observation['observation'] = state['observation']
                observation['visit'] = state['visit']
                observation['status'] = state['status']
                # Short observations are background:
                if state['duration'] < 1.5:
                    continue
                # placeholder
                observation['event'] = ''

                target_id = obs.find('apt:TargetID', ns).text
                target_name = targets[target_id]['name']
                norm_target = normalize_name(target_name)
                observation['target'] = norm_target
                observation['target_in_program'] = target_name

                observation['ra'] = targets[target_id]['ra']
                observation['dec'] = targets[target_id]['dec']
                observation['instrument'] = obs.find('apt:Instrument', ns).text
                template = obs.find('apt:Template', ns)[0]
                observation.update(_extract_instrument_template(template))

                observation['duration'] = state['duration']
                observation['date_start'] = state['date_start']
                observation['date_end'] = state['date_end']
                observation['plan_window'] = state['plan_window']
                observation['proprietary_period'] = int(proprietary_period)
                label = obs.find('apt:Label', ns)

                observation['label'] = group_label
                if label is not None:
                    label = label.text.lower()
                    observation['label'] = " : ".join([group_label, label])

                observation['special_reqs'] = reqs
                observation['phase_reqs'] = None
                if phase_reqs is not None:
                    observation['phase_reqs'] = phase_reqs.attrib
                # Add orbital-phase information when possible
                period, phase, obs_duration = _get_phase_info(observation)
                if period is None:
                    observation['period'] = None
                    observation['phase_start'] = None
                    observation['phase_duration'] = None
                else:
                    observation['period'] = period
                    observation['phase_start'] = phase
                    observation['phase_duration'] = obs_duration
                observation['event'] = guess_event_type(observation)
                observations.append(observation)

    # Cross check with NASA catalog to identify planets
    targets = [
        target for target in Catalog().targets
        if target.is_transiting
    ]
    for obs in observations:
        obs['planets'] = get_planet_letters(obs, targets)

    # Write to CSV file
    if to_csv is not None:
        fieldnames = {key for obs in observations for key in obs.keys()}
        fieldnames = []
        for obs in observations:
            for key in obs.keys():
                if key not in fieldnames:
                    fieldnames.append(key)

        with open(to_csv, mode="w", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            writer.writeheader()
            for row in observations:
                writer.writerow(row)

    return observations


def _planet_from_label(text):
    """
    Extract planet letter from label by identifying single letters
    surrounded by spaces.

    Parameters
    ----------
    text: String
        The input string.

    Returns
    -------
    list: A list of extracted single letters.
    """
    # Edge-case, more than one planet:
    if '+' in text:
        label = text.replace(' + ', '+')
        idx = label.index('+')
        planet_letters = [label[idx-1], label[idx+1]]
        if planet_letters[0].isalpha() and planet_letters[1].isalpha():
            return planet_letters

    # Regular expression to match single letters surrounded by spaces
    pattern = r'(?<=\s)[a-zA-Z](?=\s)'
    # Find all matches
    matches = re.findall(pattern, f' {text} ')
    planet_letters = np.unique(matches).tolist()
    return planet_letters


def _planet_from_period(obs, planets):
    """
    Extract planet letter from known periods.
    """
    period = obs['period']
    periods = np.array([planet.period for planet in planets])
    idx = np.argmin(np.abs(periods-period))
    planet = planets[idx].planet
    letter = get_letter(planet).strip()
    return letter


def _clean_label(label, hosts):
    """
    Clean up an observation label. Mostly to make sure that the
    planet letter is isolated when running get_planet_letters().

    Parameters
    ----------
    label: String
    hosts: 1D iterable of strings

    Returns
    -------
    label: String
    """
    if label is None:
        return ''
    label = label.lower()
    label = label.replace('776.01', '776 c')
    label = label.replace('776.02', '776 b')
    label = label.replace('836.01', '836 c')
    label = label.replace('836.02', '836 b')
    label = label.replace('175.01', '175 c')
    label = label.replace('175.02', '836 b')
    label = label.replace('.01', ' b')
    label = label.replace('.02', ' c')

    label = label.replace('_transit', ' transit')
    label = label.replace('l23', 'l 23')
    label = label.replace('l98', 'l 98')
    label = label.replace('hd106', 'hd 106')
    label = label.replace('hip67', 'hip 67')

    for host in hosts:
        label = label.replace(host.lower(), '').strip()
    return label.strip()


def get_observation_host(obs, targets):
    """
    Cross-check observation's target with catalog to find the host name

    Parameters
    ----------
    observation: Dictionary
        A dict with a JWST TSO's information
    targets: List of Target
        Catalog of targets to cross-check the observation's target name.

    Returns
    -------
    host: String
        The host name of the observation's target.
    """
    target_name = obs['target']
    name = normalize_name(target_name)

    for target in targets:
        hosts = [target.host] + [get_host(alias) for alias in target.aliases]
        if name in hosts:
            return target.host
    # If not found
    return None


def get_planet_letters(obs, targets, verbose=False):
    """
    Find the planet(s) targeted by a given JWST observation
    These are found from the observations info by looking at:
    - if the target is in a single-planet system
    - if the target- or observation-label names the planet
    - if the period matches known orbital periods in the system
    - manually identified

    Parameters
    ----------
    observation: Dictionary
        A dict with a JWST TSO's information
    targets: List of Target
        Catalog of targets to cross-check the observation's target name.

    Returns
    -------
    planet_letters: List of strings
        List of all planets (known to be) targetted by an observation.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> obs = cat.parse_program(pid=3712)

    >>> targets = [
    >>>     target for target in cat.Catalog().targets
    >>>     if target.is_transiting
    >>> ]
    >>> letters = cat.get_planet_letters(obs[0], targets)
    """
    pid = obs["pid"]
    obs_id = obs["observation"]
    visit = obs["visit"]
    info = f'{pid} {obs_id:3} {visit}  '

    # Start with the exceptions, hardcoded patching:
    if pid=='5191':
        return ['b', 'c']
    if pid=='5177':
        return ['b']
    if pid=='8739':
        return ['b']
    if pid=='2420' and obs_id=='5' and visit=='1':
        return ['c']
    if pid=='3818' and obs_id=='2' and visit=='1':
        return ['d']

    target_name = obs['target']
    # The planet is in the 'target'
    if target_name[-1].lower() == 'b' and not target_name[-2].isalpha():
        name = target_name[:-1]
        planet_letters = ['b']
        if verbose:
            print(f'{info}{target_name:15}  {planet_letters}')
        return planet_letters

    # Cross-check target with catalog:
    name = normalize_name(target_name)
    # Find host / planets
    planets = []
    for target in targets:
        aliases = [target.planet] + target.aliases
        hosts = [get_host(alias) for alias in aliases]
        if name in hosts:
            planets.append(target)

    # Single-planet systems
    if len(planets) == 1:
        planet_letters = [get_letter(planets[0].planet).strip()]
        return planet_letters

    # From label
    if len(planets) > 0:
        obs_label = obs['label']
        hosts = [planets[0].host] + [get_host(a) for a in planets[0].aliases]
        label = _clean_label(obs_label, hosts)
        planet_letters = _planet_from_label(label)
        if len(planet_letters) > 0:
            if verbose:
                print(f'{info}{repr(obs_label)}  {planet_letters}')
            return planet_letters

        # From period
        if obs['period'] is not None:
            planet_letters = [_planet_from_period(obs, planets)]
            if verbose:
                period = obs["period"]
                periods = np.round([p.period for p in planets], 1)
                print(f'{info}{name:15}{period:6.1f}  {periods}  {planet_letters}')
            return planet_letters

    print(f'Could not determine planet for {repr(name)} (PID={pid})')
    return []

